/*
 * This module is an interface to the ZORAN ZR367x0 DVD decoder chip,
 * also known as I34.
 *
 * (c) 1999-2000 David Barth <dbarth@besancon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#define __NO_VERSION__
#include <linux/module.h>
#include "mgacore.h"
#include "mgacodec.h"
#include "i34.h"

/****************************************************************************
 * LOWLEVEL DRIVER                                                          *
 ****************************************************************************/

/* main registers */
#define ZR_PARAM_ADDR	0x0
#define ZR_PARAM_DATA	0x1
#define ZR_ISR		0x2
#define ZR_IMR		0x2
#define ZR_STATUS0	0x3
#define ZR_HOSTCMD	0x3
#define ZR_STATUS1	0x4
#define ZR_BITSTREAM	0x4
#define ZR_STATUS2	0x5
#define ZR_DVP_DATA	0x6
#define ZR_DEVICE_ID	0x7
#define ZR_ADP_DATA	0x8
#define ZR_ADP_STATUS	0x8
#define ZR_SYS_CLOCK	0x9
#define ZR_NAV_DATA	0xA
#define ZR_RESERVED1	0xB
#define ZR_SUBPICBUF	0xC
#define ZR_VIDEOBUF	0xD
#define ZR_AUDIOBUF	0xE
#define ZR_RESERVED2	0xF

/* indirect registers */
#define ZR_SYSCONFIG		0x00
#define ZR_VIDCONFIG		0x01
#define ZR_SDCONFIG		0x02
#define ZR_DVPGEN1		0x03
#define ZR_DVPGEN2		0x04
#define ZR_DVPGEN3		0x05
#define ZR_COLORV		0x08
#define ZR_COLORY		0x09
#define ZR_COLORU		0x0A
#define ZR_VIDOUT		0x10
#define ZR_HTOTAL		0x11
#define ZR_VTOTAL		0x12
#define ZR_HSYNCSIZE		0x13
#define ZR_VSYNCSIZE		0x14
#define ZR_CBHSTART		0x15
#define ZR_CBVSTART		0x16
#define ZR_CBHEND		0x17
#define ZR_CBVEND		0x18
#define ZR_ACTIVESTARTX		0x19
#define ZR_ACTIVESTARTY		0x1A
#define ZR_ACTIVEENDX		0x1B
#define ZR_ACTIVEENDY		0x1C
#define ZR_ACTIVESIZEX		0x1D
#define ZR_ACTIVESIZEY		0x1E

#define ZR_CAPTIONOFFSET	0x1F
#define ZR_STARTDISPLAY 	0x20
#define ZR_OSDSWITCH		0x2D
#define ZR_DISCTYPE		0x38
#define ZR_BITSTREAMSELECT	0x40
#define ZR_PLAYBACKMODE		0x41
#define ZR_AUDSID		0x42
#define ZR_VIDSID		0x43
#define ZR_SPSID		0x44
#define ZR_AUDPORTDELAY		0x45
#define ZR_VIDPORTDELAY		0x46
#define ZR_VIDTOLERANCE		0x47
#define ZR_VIDSYNCMODE		0x48
#define ZR_SPPALETTE		0x49
#define ZR_AUTOSCALING		0x50
#define	ZR_PANSCANBASEX		0x51
#define	ZR_PANSCANBASEY		0x52
#define	ZR_PANSCANSIZEX		0x53
#define	ZR_PANSCANSIZEY		0x54

#define ZR_IMAGESTARTX		0x57
#define ZR_IMAGESTARTY		0x58
#define ZR_IMAGESIZEX		0x59
#define ZR_IMAGESIZEY		0x5A
#define ZR_IMAGEENDX		0x5B
#define ZR_IMAGEENDY		0x5C

#define ZR_SCALERATIO		0x5E
#define ZR_AUTOPANSCAN		0x68
#define ZR_PANSCANOFFSETX	0x69
#define ZR_PANSCANOFFSETY	0x6A
#define ZR_BACKGROUNDSWITCH	0x6B
#define ZR_SPSWITCH		0x6C

#define ZR_CAPTIONSWITCH	0x6F

#define ZR_INITFILE_1		0xF0
#define ZR_INITFILE_2		0xF1
#define ZR_INITFILE_3		0xF2
#define ZR_INITFILE_4		0xF3
#define ZR_MICROCODE		0xFD

/* host commands */
#define I34_CMD_START	0x0000
#define I34_CMD_PAUSE	0x0100
#define I34_CMD_STEP	0x0300
#define I34_CMD_SLOW	0x0500
#define I34_CMD_END    	0x0700
#define I34_CMD_FAST	0x0800
#define I34_CMD_CONT	0x0900
#define I34_CMD_SEQEND	0x0E00

struct video_standard_s
{
	u16 VidFPS;
	u16 FieldSel;
	u16 HTotal;
	u16 VTotal;
	u16 HSyncSize;
	u16 VSyncSize;
	u16 ActiveStartX;
	u16 ActiveStartY;
	u16 ActiveEndX;
	u16 ActiveEndY;
	u16 ActiveSizeX;
	u16 ActiveSizeY;
	u16 VidConfig;
};

struct composite_blank_s
{
	u16 CBHStart;
	u16 CBVStart;
	u16 CBHEnd;
	u16 CBVEnd;
};

struct serial_port_config_s
{
	/* see p. 226 for a definition of the structure */
	u8 cpb;
	u8 cpa;
	u8 frb;
	u8 fra;
	u8 err;
	u8 isp;
	u8 osp;
	u8 spo;
	u8 inw;
	u8 outw;
	u8 spbs;
	u8 fmb;
	u8 fma;
};

struct mga_dev_s;

struct i34_dev_s
{
/* codec configuration section */
	/* WARNING: not all zoran 367x0 registers are 2 bytes long
	   but it's enough for those we're interested in reading back
	*/
	u16 shadow_register[0xff];
	u32 CodeBurstLen;
	u32 init_file_loaded;
	struct mgadvd_status_s status_summary;
	struct mga_codec_s *codec_interface;
	struct serial_port_config_s serial;

/* video configuration section */
	u32 dvp_microcode_loaded;
	u32 VideoStandard; 
	u32 SubPicturePalette[16];
	u32 BgSwitch;
	u32 AutoScaling;
	u32 AutoPanScan;
	u32 ScaleRatio;
	u32 VidEntry;
	u32 LastPic;
	u32 VidFloat;
	u32 SpSwitch;
	u32 CaptionSwitch;
	u32 video_tolerance;
	u32 video_sync_mode;
	u32 video_sid;
	u32 subpic_sid;
	struct video_standard_s *vs;
	struct composite_blank_s cbs;
	struct video_key ckey;

/* audio configuration section */
	u32 adp_microcode_loaded;
	u32 OutputSamplingRate;
	u32 audio_gain;
	u32 amclk;
	u32 spbs;
	u32 audio_tolerance;
	u32 audio_sync_mode;
	u32 audio_sid;

/* CSS configuration section */
#ifdef CSS_SUPPORT
	dvd_authinfo cssauth;
#endif
};

static struct i34_dev_s i34;

static u8 dvp_dvd[] = {
#include "340edvd.inc"	/* DVD playback */
};
static u8 dvp_vcd[] = {
#include "340ev4p.inc"	/* VCD playback */
};
static u8 dvp_oth[] = {
#include "340eoth.inc"	/* MPEG-1,2 video playback */
};
static u8 adp_ac3[] = {
#include "340eac3.inc"	/* Dolby 2.0, 2 channel audio, soundcard CD in */
};
static u8 adp_ac36[] = {
#include "340eac36.inc" /* Dolby 5.1, 6 channel audio, Marvel S/PDIF out */
};
static u8 adp_lpcm[] = {
#include "340elpcm.inc" /* DVD LPCM, 8 channel audio */
};
static u8 adp_mpeg[] = {
#include "340empeg.inc" /* MPEG-1,2 Layer II, 2 channel audio */
};
static u8 adp_cdda[] = {
#include "340ecdda.inc" /* CD Digital Audio, 2 and 4 channel audio */
};

u16 i34_read(u8 reg)
{
	mgacodec_push_readw(i34.codec_interface, reg);
	mgacodec_exec_commands(i34.codec_interface);
	return mgacodec_get_word(i34.codec_interface);
}

void i34_write(u8 reg, u16 data)
{
	mgacodec_push_writew(i34.codec_interface, reg, data);
	mgacodec_exec_commands(i34.codec_interface);
}

u16 i34_read_parameter(u8 addr)
{
	/* see table 32 p. 79 for the protocol to access the indirect address space
	   addr. in the l.s. bits, m.s. bits must be 0
	*/
	mgacodec_push_writew(i34.codec_interface, ZR_PARAM_ADDR, addr & 0x00FF);
	mgacodec_push_readw(i34.codec_interface, ZR_PARAM_DATA);
	mgacodec_exec_commands(i34.codec_interface);
	/* we need shadow registers, because most of them are write only...
	 especially when dealing with and/or operations
	*/
	i34.shadow_register[addr] = mgacodec_get_word(i34.codec_interface);
	return i34.shadow_register[addr];
}

void i34_write_parameter(u8 addr, u16 data)
{
	i34.shadow_register[addr] = data;
	mgacodec_push_writew(i34.codec_interface, ZR_PARAM_ADDR, addr & 0x00FF);
	mgacodec_push_writew(i34.codec_interface, ZR_PARAM_DATA, data);
	mgacodec_exec_commands(i34.codec_interface);
}

/* change some bits of a parameter register.
   this could also be named i34_parameter_and_or...
*/
void i34_change_parameter(u16 addr, u16 or, u16 and)
{
	/* FIXME: I should make sure that we have previously _written_ to this register,
	          or we modify an incoherent value...
	*/
	u16 value = (i34.shadow_register[addr] & ~and) | or;
	i34_write_parameter(addr, value);
}
	
int i34_write_seq(u8 addr, u16* buffer, u32 size_in_words)
{
	u32 i;

	if (buffer == NULL) return 0;

	for (i=0; i<size_in_words; i++) {
		mgacodec_push_writew(i34.codec_interface, addr, *buffer++);
		mgacodec_exec_commands(i34.codec_interface);
	}
	return 1;
}

int i34_write_parameter_seq(u8 addr, u16* buffer, u32 size_in_words)
{
	if (buffer == NULL) return 0;

	/* multiple read/writes without modifying the Parameter Address Register */
	mgacodec_push_writew(i34.codec_interface, ZR_PARAM_ADDR, addr & 0x00FF);
	return i34_write_seq(ZR_PARAM_DATA, buffer, size_in_words);
}

/* read the System Clock */
u32 i34_read_sclk(void)
{	
	u16 a, b;
	a = i34_read(ZR_SYS_CLOCK);
	b = i34_read(ZR_SYS_CLOCK);
	return (a<<16 | b);
}

/* wait 'delay' microseconds */
int i34_wait(int delay)
{
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(delay * HZ/1000000);
	return 1;
}

int i34_reset_pin(void)
{
	struct mga_codec_s *codec;

	codec = i34.codec_interface;

	/* program the MISC<2> and DDC<2> lines
	   DDC<2> is connected to HRDY according to the G100 specs	   
	*/
	mgacodec_set_misc2(codec, 1);
	mgacodec_set_ddc2(codec, 1);

	mgacodec_reset(codec);

	mgacodec_and_or_miscctl(codec, 0x00, 0x00);
	mgacodec_and_or_miscctl(codec, 0x00, 0x03);
	mgacodec_and_or_miscctl(codec, 0x00, 0x03);
  
	mgacodec_and_or_miscctl(codec, ~0x0f, 0x07);
	mdelay(10);
	mgacodec_and_or_miscctl(codec, ~0x0f, 0x03);
	mdelay(1);
	mgacodec_and_or_miscctl(codec, ~0x0f, 0x07);
  
	mgacodec_and_or_miscctl(codec, ~0x0f, 0x07);
	mdelay(100);
	mgacodec_and_or_miscctl(codec, ~0x0f, 0x0f);
	mdelay(1);
	mgacodec_and_or_miscctl(codec, ~0x0f, 0x07);
	mdelay(200);

	mgacodec_and_or_miscctl(codec, 0x0f, 0xf9);


	/* codec_test_lines(mga, 1); */
	return 1;
}

int i34_wait_for_pll(void)
{
	u16 i, isr;

	for (i=0; i<5000; i++) {
		isr = i34_read(ZR_ISR);
		if ((isr & 0x1800) == 0x1800) {
			i34_write_parameter(ZR_INITFILE_4, 0xA0);
			i34_write_parameter(ZR_INITFILE_4, 0x88);
			return 1;
		}
		i34_wait(100);
	}
	dprintk("PLL's NOT locked ---> 0x%4.4x\n", isr);
	return 0;
}

int i34_reset(int flag)
{
	dprintk("i34: Reset\n");

	if (! i34_reset_pin() ) return 0;

	i34_wait(5000);

	if (flag) return 1;

	return i34_wait_for_pll();
}

int i34_state(int *s)
{
	*s = (i34_read(ZR_STATUS0) >> 10);
	return 1;
}

/* NEW ADP INTERFACE */

/* status bits */
#define ADPRDY		1
#define ADPEMPTY	2

/* ADP commands */
#define ADP_READ	0x00
#define ADP_NOP		0x80
#define ADP_VER		0x81
#define ADP_CFG		0x82

/* ADP audio microcode function setup */
#define ADP_PINK	0x83
#define ADP_LPCM	0x84
#define ADP_AC3		0x85
#define ADP_PCM		0x86
#define ADP_MPEG	0x87

/* ADP playback control */
#define ADP_UNMUTE      0x89
#define ADP_PLAY        0x8A
#define ADP_MUTE        0x8B
#define ADP_STOP        0x8C
#define ADP_STOPF       0x8D

/* ADP misc configuration */
#define ADP_STAT        0x8E
#define ADP_BOOT        0x90
#define ADP_SETIO       0x92
#define ADP_PEEK        0x94
#define ADP_SPDIFCS     0x95
#define ADP_PARAM       0x96
#define ADP_PLLTAB      0x98
#define ADP_PLLCFG      0x99

/* ADP return codes */
#define ADP_NO_ERROR		0x80
#define ADP_OPCODE_ERROR	0x81
#define ADP_EXT_ERROR		0x82
#define ADP_NOT_READY_ERROR	0x83
#define ADP_RESERVED_ERROR	0x84
#define ADP_READY_FOR_NEW_CMD	0x85

u8 adp_buffer[64];

/* clean the pending bytes waiting in the ADP Data Register
   used to start from a clean command/read pipe, because of
   the offset between commands sent and response byte
   (off by 2)
*/
int adp_read(u8 *data)
{
	u16 status0;

	/* see table 48 p. 115 */
	status0 = i34_read(ZR_STATUS0);
	if (!(status0 & ADPRDY)) {
		i34_wait(200);
		status0 = i34_read(ZR_STATUS0);
		if (!(status0 & ADPRDY)) {
			dprintk("i34: Unable to read ADP register (!ADPRDY)\n");
			return 0;
		}
	}
	
	*data = (i34_read(ZR_ADP_DATA) & 0xFF);
	
	return 1;
}

int adp_write(u8 data)
{
	u16 status0;

	status0 = i34_read(ZR_STATUS0);
	if (!(status0 & ADPEMPTY)) {
		i34_wait(200);
		status0 = i34_read(ZR_STATUS0);
		if (!(status0 & ADPEMPTY)) {
			dprintk("i34: Unable to write ADP register (!ADPEMPTY)\n");
			return 0;
		}
	}

	i34_write(ZR_ADP_DATA, (data & 0x00FF));

	return 1;
}

int adp_clean(void)
{
	u16 status0;
	u8 istatus1, istatus2;
	int i;

	/* empty ADP buffered answers */
	adp_write(ADP_NOP);
	for (i=0; i<1000; i++) {
		status0 = i34_read(ZR_STATUS0);
		if (status0 & ADPRDY) {
			adp_write(ADP_NOP);
			adp_read(&istatus1);
			adp_write(ADP_NOP);
			adp_read(&istatus2);
			dprintk("i34: adp_cleaned NOP->%#02x, NOP->%#02x\n", istatus1, istatus2);
			return 1;
		}
	}
	dprintk("i34: Unable to clean ADP (!ADPRDY)\n");
	return 0;
}

int adp_handle_error(u8 opcode_or_ext, u8 error_code, u8 awaited)
{
	/* no need to go further if the adp returned what we were waiting for */
	if (error_code == awaited) return 1;

	switch(error_code) {
	case ADP_NO_ERROR:
		return 1;
		break;
	case ADP_OPCODE_ERROR:
		dprintk("i34: ADP opcode error (opcode %#2x)\n", opcode_or_ext);
		break;
	case ADP_EXT_ERROR:
		dprintk("i34: ADP EXT param error (EXT %#2x)\n", opcode_or_ext);
		break;
	case ADP_NOT_READY_ERROR:
		dprintk("i34: ADP not ready for new command (opcode %#2x)\n", opcode_or_ext);
		break;
	case ADP_RESERVED_ERROR:
		dprintk("i34: ADP reserved error (opcode %#2x)\n", opcode_or_ext);
	case ADP_READY_FOR_NEW_CMD:
		/* FIXME: is it really an error ? */
		dprintk("i34: ADP ready for new command (opcode %#2x)\n", opcode_or_ext);
	default:
		dprintk("i34: ADP count mismatch (%d ack for %d supposed after opcode %#2x)\n",
			error_code, awaited, opcode_or_ext);
	}
	adp_clean();
	return 0;
}

int adp_simple_command(u8 opcode)
{
	u8 dummy, istatus;

	/* same as below but we don't need to check for a parameter count 
	   during the write sequence: just make sure that we get ADP_NO_ERROR
	   at the end
	*/

	adp_write(opcode);
	adp_read(&dummy);
	adp_write(ADP_READ);
	adp_read(&istatus);
	if (! adp_handle_error(opcode, istatus, ADP_NO_ERROR) ) return 0;

	return 1;
}

int adp_exec_command(u8 opcode, u8 *buffer, int param_len, int resp_len)
{
	u8 dummy, count, istatus;
	int index = 0;

	/* see p. 115 & 204 */	
	adp_write(opcode);		/* send opcode */
	adp_read(&dummy);
	while (param_len > 0) {
		adp_write(buffer[index++]);
		adp_read(&count);		/* count of bytes left to send (+1) */
		/* check if number of awaited parameters matches the param_len */
		if (! adp_handle_error(opcode, count, param_len) ) return 0;
		param_len--;
	}
	/* just before reading the response bytes the count MUST be 1 */
	if (! adp_handle_error(opcode, count, 1) ) return 0;
	index = 0;
	while (resp_len > 0) {
		adp_write(ADP_READ);
		adp_read(&buffer[index++]);
		resp_len--;
	}
	adp_write(ADP_NOP);
	adp_read(&istatus);		/* reading ISTATUS */

	if (! adp_handle_error(opcode, istatus, ADP_NO_ERROR) ) return 0;

	return 1;
}

/* configure the GPAI/O pins (see p. 231) */
int adp_config_gpaio(u8 iocmask, u8 iocbits)
{
	adp_buffer[0] = iocmask;
	adp_buffer[1] = iocbits;
	adp_buffer[2] = 0;
	adp_buffer[3] = 0;
	return adp_exec_command(ADP_SETIO, adp_buffer, 4, 2);
}

int adp_set_gpaio(u8 iomask, u8 iobits)
{
	adp_buffer[0] = 0;
	adp_buffer[1] = 0;
	adp_buffer[2] = iomask;
	adp_buffer[3] = iobits;
	return adp_exec_command(ADP_SETIO, adp_buffer, 4, 2);
}

int adp_rom_version(void)
{
	u8 istatus, dummy;

	return 1;

	/* I'm not sure of the commands below, still this ADP weird "off by 2" bug */

	adp_write(ADP_VER);
	adp_read(&dummy);
	adp_write(ADP_READ);
	adp_read(&adp_buffer[0]);
	adp_write(ADP_READ);
	adp_read(&adp_buffer[1]);
	adp_write(ADP_READ);
	adp_read(&adp_buffer[2]);
	adp_write(ADP_READ);
	adp_read(&adp_buffer[3]);
	adp_write(ADP_READ);
	adp_read(&istatus);
	if (istatus != ADP_NO_ERROR) {
		dprintk("i34: adp_rom_version() error %#02x\n", istatus);
		return 0;
	}
	dprintk("i34: ADP ROM Version: %#8.8x\n", ((int*)adp_buffer)[0]);
	return 1;
}


struct pll_tab_s
{
	u8 sr;
	u16 audd, audm;
	u8 dspd, dspm;
};

/* GCLK1 = 27MHz
   fs = 32, 44.1, 48 or 96 KHz
   AMCLK = {128, 192, 256 or 384} * fs
   AMCLK = GCLK1 * AUDM / AUDD

   example : 48KHz
   AMCLK = 256 * 48Khz = 0x3000
   AMCLK = 27000 * 0x0200 / 0x465= 0x3000
*/

struct pll_tab_s pll_tab[] = {
	{ 2, 0x0d2f, 0x0400, 0, 0 },
	{ 1, 0x0753, 0x0310, 0, 0 },
	{ 0, 0x0465, 0x0200, 0, 0 },
	{ 3, 0, 0, 0, 0 },

	{ 2, 0x0465, 0x0200, 0, 0 },
	{ 1, 0x0271, 0x0188, 0, 0 },
	{ 0, 0x0177, 0x0100, 0, 0 },
	{ 3, 0, 0, 0, 0, },
};

char srtab[] = {48, 44, 32, 96};

/* compute plltab value for current bitrate and output frequency
   gclkhz = 27000  (27000kHz == 27MHz)
   AUDD = 1125 (given)
   AUDM = kbitrate * kilohertz / gclkhz
   DSPM = 3 (given)
   DSPD = 1 (given)

   example:  768 = 384 * 48 * 1125 / 27000
             AUDM = 768, AUDD = 1125
*/

int adp_plltab(void)
{
	u32 audm, audd, dspm, dspd, kbps, khz, gclkhz;

	gclkhz = 27000;	/* 27000kHz for gclock (27MHz) */
	khz = 48;	/* FIXME: defaulting to 48khz output for now */
	dspm = 3;
	dspd = 1;
	audd = 1125;

	i34.amclk = 384;
	i34.spbs = 3;

	kbps = i34.amclk;

	/* khz not as accurate as hz (at 44.1khz), but keeps math 32 bit 
	   and since we're staying at 48khz (always?) */

	audm = (kbps * khz * audd) / gclkhz; 

	adp_buffer[0] = (0 << 5) | ( (audd & 0xFF00) >> 8);
	adp_buffer[1] = audd & 0x00FF;
	adp_buffer[2] = (audm & 0xFF00) >> 8;
	adp_buffer[3] = audm & 0x00FF;
	adp_buffer[4] = dspd;
	adp_buffer[5] = dspm;
	
	dprintk("i34: PLLTAB %dkbps %dkhz %d/%d\n",kbps,khz,audm,audd);

	return adp_exec_command(ADP_PLLTAB, adp_buffer, 6, 0);
}


int adp_plltab_o(void)
{

	struct pll_tab_s *table;
	int i, set = 0;

	/* see p. 36 & 228 */
	/* default values after reset :
	   DSPM = 3
	   DSPD = 1
	*/

	set = (i34.amclk == 384) ? 1 : 0;

	for (i=0; i<3; i++) {
		table = &pll_tab[i + set * 4];
		adp_buffer[0] = (table->sr << 5) | ((table->audd & 0xFF00)>>8);
		adp_buffer[1] = (table->audd & 0x00FF);
		adp_buffer[2] = (table->audm & 0xFF00) >> 8;
		adp_buffer[3] = (table->audm & 0x00FF);
		adp_buffer[4] = table->dspd;
		adp_buffer[5] = table->dspm;
		adp_exec_command(ADP_PLLTAB, adp_buffer, 6, 0);
	}
	return 1;
}

int adp_pllcfg(u8 sr)
{
	int pa, pd;

	/* see p. 229 */
	adp_plltab();
	dprintk("i34: PLLCFG(%d)\n", sr);
	adp_buffer[0] = (sr & 7)<<5;
	/* relock AMCLK unless SR=111b (p. 38) */
	if (sr != 7)
		adp_buffer[0] |= 2;

	if (! adp_exec_command(ADP_PLLCFG, adp_buffer, 1, 1) ) {
		dprintk("i34: error during PLLCFG\n");
		/* don't exit now: try to get a stable clock first !!! */
		i34_wait(50);
		i34_wait_for_pll();
		return 0;
	}
	pa = adp_buffer[0] & 0x02;
	pd = adp_buffer[0] & 0x01;
	dprintk("i34: AMCLK %s; PCLK %s\n",
		pa ? "locked" : "not locked",
		pd ? "locked" : "not locked");
	
	i34_wait(50);
	i34_wait_for_pll();

	return 1;
}

int adp_config_serial(struct serial_port_config_s *config)
{
	adp_buffer[0] = 0xc0;
	adp_buffer[1] = 0x18;
	adp_buffer[2] = (config->fra & 7) | (config->frb & 7)<<3 | (config->cpa & 1)<<6 | (config->cpb & 1)<<7;
	adp_buffer[3] = (config->osp & 1) | (config->isp & 1)<<1 | (config->err & 1)<<4;
	adp_buffer[4] = (config->outw & 3)<<1 | (config->inw & 3)<<3 | (config->spo & 1)<<5;
	adp_buffer[5] = 0;

	/* we assume AMCLK = 256 (see table 29 p. 73) */
	/*
	switch (i34.amclk) {
	case 256:
		if (config->frb == 0)
			config->spbs = 4;
		else 
			config->spbs = 2;
		break;
	case 384:
		if (config->frb == 0)
			config->spbs = 6;
		else if (config->frb == 1)
			config->spbs = 3;
		else
			config->spbs = 4;
		break;
	default:
		dprintk("i34: invalid AMCLK\n");
		return 0;
	}
*/
	config->spbs = i34.spbs;

	adp_buffer[6] = config->spbs;
	adp_buffer[7] = (config->fma & 7) | (config->fmb & 7)<<3;

	if (! adp_exec_command(ADP_CFG, adp_buffer, 8, 0) ) return 0;

	return 1;
}

int adp_config_spdif(void)
{
	switch (i34.OutputSamplingRate) {
	case 0:
		adp_buffer[0] = 2;		/* 48KHz (see p. 230) */
		break;
	case 1:
		adp_buffer[0] = 0;		/* 44.1KHz (see p. 230) */
		break;
	case 2:
		adp_buffer[0] = 3;		/* 32KHz (see p. 230) */
		break;
	default:
		dprintk("i34: unknown sampling rate\n");
		return 0;
	}
	adp_buffer[1] = 0;		/* IEC 958 category mode (???) */
	adp_buffer[2] = 0;		/* PCM encoded data */
	adp_buffer[3] = 0;
	return adp_exec_command(ADP_SPDIFCS, adp_buffer, 4, 0);
}

int adp_audio_tolerance(u16 tolerance)
{
	adp_buffer[0] = 0x02;
	adp_buffer[1] = tolerance >> 8;
	adp_buffer[2] = tolerance;
	return adp_exec_command(ADP_PARAM, adp_buffer, 3, 0);
}

int adp_audio_sync(u8 sync)
{
	adp_buffer[0] = 0x03;
	adp_buffer[1] = 0;
	adp_buffer[2] = sync & 3;
	return adp_exec_command(ADP_PARAM, adp_buffer, 3, 0);	
}

int i34_audio_sync(int sync)
{
	i34.audio_sync_mode = sync & 3;
	return adp_audio_sync(i34.audio_sync_mode);
}

int i34_video_sync(int sync)
{
	i34.video_sync_mode = sync & 1;
	return 1;
}

int adp_set_function(void)
{
	int function_to_call;

        switch (i34.adp_microcode_loaded) {
	case ADP_MICROCODE_AC36:
        case ADP_MICROCODE_AC3:
/* see p. 206 */
/*
 summary of AC3 (DVD AC3 audio) function bits current settings:
 no ProLogic, Compression mode is line mode (no effect?), no Dual Mono Mode
 no Surround Filter, no Auto Balance, no Bass redirection, no SubWoofer,
  2/0 Output speaker configuration with downmix
  (can try w/ or w/o downmix or other spkr cfgs, tests indicate no effect?)
 No Center Delay, no Surround Delay (16 enables 3D, no effect?)
 High & Low Dynamic Range set to maximum range for loud and soft sounds
  (experiment with this value. max seems to make all sounds less distorted) 
 Karaoke disabled, no Repeat Count (repeats sound like bad synth jazz)
 PCM Scale Factor to max audio gain
  (can change this during playback w/ ADP PARAM EXT 00?)
*/
		/*              PRLG    COMP    DMM 			*/
		adp_buffer[0] = (0<<6) |(1<<2) |0;
		/*              SF      AB      BCFG    SW      OCFG 	*/
		adp_buffer[1] = (1<<7) |(1<<6) |(0<<4) |(0<<3) |2;
		/*              CDLY    SRDLY 				*/
		adp_buffer[2] = (0<<5) |0;
		/*              HDYNRNG (00-7F range) 			*/
		adp_buffer[3] = 0x7F;
		/*              LDYNRNG (00-7F range) 			*/
		adp_buffer[4] = 0x7F;
		/*              KAR     RPC 				*/
		adp_buffer[5] = (0<<3) |0;
		/*              PCMSF (ms byte 00-7F range)		*/
		adp_buffer[6] = 0x7F;
		/*              PCMSF (ls byte 00-FF range)		*/
		adp_buffer[7] = 0xFF;
                function_to_call = ADP_AC3;
                /* AC-3 audio stream 0 (see p. 128) */
                i34.audio_sid = 0x0080;   /* FIXME: should be + cmdparm id */
                break;

        case ADP_MICROCODE_MPEG:
/* see p. 213 */
/*
 summary of MPEG (DVD MPEG audio) function bits current settings:
 similar to AC3 function but without: COMP, HDYNRNG, LDYNRNG, KAR, RPC
 no ProLogic, no Dual Mono Mode
 no Surround Filter, no Auto Balance, no Bass redirection, no SubWoofer,
  2/0 Output speaker configuration without downmix
 No Center Delay, no Surround Delay (16 enables 3D, no effect?)
 PCM Scale Factor to max audio gain
  (can change this during playback w/ ADP PARAM EXT 00?)
*/
		/*              PRLG    DMM			     */                               
		adp_buffer[0] = (0<<6) |0;
		/*              SF      AB      BCFG    SW      OCFG */
		adp_buffer[1] = (1<<7) |(1<<6) |(0<<4) |(0<<3) |2;
		/*              CDLY    SRDLY			     */
		adp_buffer[2] = (0<<5) |0;
		/*              not used set to 0		     */
		adp_buffer[3] = 0x00;
		/*              not used set to 0		     */
		adp_buffer[4] = 0x00;
		/*              not used set to 0		     */
		adp_buffer[5] = 0x00;
		/*              PCMSF (ms byte 00-7F range)	     */
		adp_buffer[6] = 0x7F;
		/*              PCMSF (ls byte 00-FF range)	     */
		adp_buffer[7] = 0xFF;
		function_to_call = ADP_MPEG;
		/* MPEG audio stream 0 (see p. 128) */
		i34.audio_sid = 0x00C0; /* FIXME: should be + cmdparm id */
		break;
		                
                

        case ADP_MICROCODE_LPCM:
/* see p. 211 */
/*
 summary of LPCM (DVD PCM audio) function bits current settings:
 similar to AC3 function but without: COMP, DMM, HDYNRNG, LDYNRNG, KAR, RPC
 added functions are DE, DRC(similar to H&LDYNRNG) and CMX
 no ProLogic, no De-Emphasis (dont want it fiddling with GPAIO0 bit?)
 no Surround Filter (Noise Reduction and Low Pass(?), experiment with these)
  no Auto Balance,  no Bass redirection, no SubWoofer,
  2/0 Output Configuration without downmix
 No Center Delay, no Surround Delay (16 enables 3D, no effect?)
 Dynamic Range Control to max (experiment with this)
 Channel Mixing uses no default OCFG=2 downmix table
 PCM Scale Factor to max audio gain
  (can change this during playback w/ ADP PARAM EXT 00?)
*/
		/*              PRLG    DE			     */
		adp_buffer[0] = (0<<6) |(0<<2);
		/*              SF      AB      BCFG    SW      OCFG */
		adp_buffer[1] = (1<<7) |(1<<6) |(0<<4) |(0<<3) |2;
		/*              CDLY    SRDLY			     */
		adp_buffer[2] = (0<<5) |0;
		/*              DRC				     */
		adp_buffer[3] = 0x00;
		/*              not used, set to 0		     */
		adp_buffer[4] = 0x00;
		/*              CMX				     */
		adp_buffer[5] = (0<<7) |0;
		/*              PCMSF (ms byte 00-7F range)	     */
		adp_buffer[6] = 0x7F;
		/*              PCMSF (ls byte 00-FF range)	     */
		adp_buffer[7] = 0xFF;
		function_to_call = ADP_LPCM;
		/* LPCM audio stream 0 (see p. 128) */
		i34.audio_sid = 0x00A0; /* FIXME: should be + cmdparm id */
		break;

	default:
		dprintk("i34: invalid ADP microcode selected for function set\n");
		return 0;
	}

	return adp_exec_command(function_to_call, adp_buffer, 8, 0);
}

int adp_audio_gain(u32 gain)
{
	/* see p. 221 */
	gain &= 0x0007FFFF;
	adp_buffer[0] = 0x00;	/* means PARAM EXT = ADP Command - Volume Control */
	adp_buffer[1] = 0x07;
	adp_buffer[2] = 0xFF;
	adp_buffer[3] = 0xFF;
	return adp_exec_command(ADP_PARAM, adp_buffer, 3, 0);
}

int adp_decoder_status(void)
{
	int i;
	u8 count, istatus;

	/* see p. 235 */
	adp_write(ADP_STAT);
	adp_read(&count);		/* count of bytes left to read */

	for (i=0; i<count; i++) {
		adp_write(ADP_READ);
		adp_read(&adp_buffer[i]);
	}
	adp_write(ADP_READ);
	adp_read(&istatus);		/* reading ISTATUS */
	if (istatus != ADP_NO_ERROR) {
		dprintk("i34: adp_decoder_status error %#02x\n", istatus);
		return 0;
	}

	return 1;
}

/* CSS sub-system */
#ifdef CSS_SUPPORT

int css_read_key(u8 *buffer, int length)
{
	int i;

	for (i=0; i<length; i++) {
		buffer[i] = i34_read_parameter(0xA1);
	}
	return 1;
}

int css_write_key(u8 *buffer, int length)
{
	int i;

	for (i=0; i<length; i++) {
		i34_write_parameter(0xA2, buffer[i]);
	}
	dprintk("i34: css status %#04x\n", i34_read_parameter(0xA1));
	return 1;
}

int css_wait(int a, int b, int *c)
{
	int i;

	for (i=0; i<1000; i++) {
		*c = i34_read_parameter(0xA1);
		i34_wait(10);
		if ((*c & a) == b)
			return 1;
	}
	return 0;
}

int css_select(int a)
{
	if ((i34_read_parameter(0xA1) & 0xC0) != 0x80)
		return 0;
	/* i34_change_parameter(0xA0... */
	return 1;
}

int css_write_disk_key(dvd_authinfo auth)
{
	return 1;
}

int css_write_title_key(dvd_authinfo auth)
{
	return 1;
}

int css_read_chal_key(dvd_authinfo auth)
{
	return 1;
}

int css_write_chal_key(dvd_authinfo auth)
{
	return 1;
}

int css_read_resp_key(dvd_authinfo auth)
{
	return 1;
}

int css_write_resp_key(dvd_authinfo auth)
{
	return 1;
}

#endif


#define HCREADY (1<<15)

int i34_host_command(u16 cmd)
{
	u16 status1;

	/* p. 100 & 147 ZR HB: check that the host is ready to receive a new command */
	status1 = i34_read(ZR_STATUS1);
	if (status1 & HCREADY) {
		i34_write(ZR_HOSTCMD, cmd);
	} else {
		dprintk("i34: host is not ready to receive commands (STATUS1=%#4.4x)\n", status1);
		return 0;
	}

	return 1;
}

int i34_send_command(u16 opcode, u8 parameter)
{
	int cstate;
	int i;
	int res;

	if (opcode == I34_CMD_START) {
		/* per p. 150, it is not advised to send START while the codec is already
		   in playback mode */
		i34_state(&cstate);
		/* if init_display or idle */
		if ((cstate == 0x03) || (cstate == 0x30)) {
			res = i34_host_command(opcode | parameter);
			i34_wait(50);
			return res;
		} else {
			dprintk("i34: incorrect state for a START command (c-state = %#2.2x)\n", cstate);
			return 0;
		}
	} else if (opcode == I34_CMD_END) {
		i34_host_command(opcode | parameter);
		/* it takes a while to bring the card down to the idle state */
		for (i=0; i<1000; i++) {
			i34_state(&cstate);
			if (cstate == 0x30) return 1;
			i34_wait(50);
		}	
		dprintk("i34: no idle state after end_playback...(c-state=%#02x)\n", cstate);
		return 0;
	}
	return i34_host_command(opcode | parameter);
}

int i34_detect(void *codec_interface)
{
	u32 id;

	i34.codec_interface = codec_interface;

	i34_reset(0);
	i34_reset(0);
	id = i34_read(ZR_DEVICE_ID);
	/* we should obtain: 0x34 m.s. bits, rev in l.s. bits p. 112 ZR HB */
	/* warning: this test is arch. dependent ! */
#ifndef __LITTLE_ENDIAN
#error "This driver does not work with big endian architectures: contact dbarth@besancon.net if you wan't to help"
#else
	if (((id>>8) & 0xFF) == 0x34) {
		dprintk("i34: Found ZR367x0 rev. 0x%2.2x, good !\n", id & 0xFF);
		return 1;
	} else {
		dprintk("i34: no chipset found (codec response: 0x%4.4x)\n", id);
		return 0;
	}
#endif
}

int i34_color_key(struct video_key *ck)
{
	i34_write_parameter(ZR_COLORV, ck->key[0]);
	i34_write_parameter(ZR_COLORU, ck->key[1]);
	i34_write_parameter(ZR_COLORY, ck->key[2]);
	return 1;
}

static struct video_standard_s ntsc1 = {
	0, 0, 858, 525,	64, 6, 136, 14,	856, 260, 720, 240, 0x4006
};

static struct video_standard_s ntsc2 = {
	0, 0, 858, 525,	64, 6, 136, 16, 856, 352, 720, 240, 0x4006
};

static struct video_standard_s pal1 = {
	1, 0, 864, 625, 64, 6, 142, 14, 860, 310, 720, 288, 0x4006
};

static struct video_standard_s pal2 = {
	1, 0, 864, 625, 64, 6, 142, 16, 862, 310, 720, 288, 0x4006
};

int i34_video_standard(int standard)
{
	int cstate;

	dprintk("i34: i34_video_standard(%d)\n", standard);
	switch(standard) {
	case VIDEO_MODE_NTSC:
		i34.vs = &ntsc1;
		break;
	case I34_NTSC_ALT:
		i34.vs = &ntsc2;
		break;
	case VIDEO_MODE_PAL:
		i34.vs = &pal1;
		break;
	case I34_PAL_ALT:
		i34.vs = &pal2;
		break;
	default:
		i34.vs = &ntsc1;
		break;
    	}

	if (! i34_state(&cstate) ) return 0;

	/* if init_pclk, init_display or idle */
	if ((cstate == 0x03) || (cstate == 0x30) || (cstate == 0x02)) {
		i34_write_parameter(ZR_VIDOUT, i34.vs->FieldSel | i34.vs->VidFPS);
		i34_write_parameter(ZR_HTOTAL, i34.vs->HTotal);
		i34_write_parameter(ZR_VTOTAL, i34.vs->VTotal);
		i34_write_parameter(ZR_HSYNCSIZE, i34.vs->HSyncSize);
		i34_write_parameter(ZR_VSYNCSIZE, i34.vs->VSyncSize);
		i34_write_parameter(ZR_ACTIVESTARTX, i34.vs->ActiveStartX);
		i34_write_parameter(ZR_ACTIVESTARTY, i34.vs->ActiveStartY);
		i34_write_parameter(ZR_ACTIVEENDX, i34.vs->ActiveEndX);
		i34_write_parameter(ZR_ACTIVEENDY, i34.vs->ActiveEndY);
		i34_write_parameter(ZR_ACTIVESIZEX, i34.vs->ActiveSizeX);
		i34_write_parameter(ZR_ACTIVESIZEY, i34.vs->ActiveSizeY);

		if (cstate == 0x02) {
			/* per p. 80 VidConfig should only be configured at init_pclk state */
			i34_write_parameter(ZR_VIDCONFIG, i34.vs->VidConfig);
		}
		return 1;
	} else {
		dprintk("i34: incorrect state for video configuration (c-state=%#2.2x)\n", cstate);
		return 0;
	}
}

int i34_composite_blank_sync(struct composite_blank_s *cb)
{
	int cstate;

	if (! i34_state(&cstate) ) return 0;

	/* if init_pclk, init_display or idle */
	if ((cstate == 0x03) || (cstate == 0x30) || (cstate == 0x02)) {
		i34_write_parameter(ZR_CBHSTART, cb->CBHStart);
		i34_write_parameter(ZR_CBVSTART, cb->CBVStart);
		i34_write_parameter(ZR_CBHEND, cb->CBHEnd);
		i34_write_parameter(ZR_CBVEND, cb->CBVEnd);
		return 1;
	} else {
		dprintk("i34: incorrect state for composite blank sync configuration (c-state=%#2.2x)\n", cstate);
		return 0;
	}
}


int i34_load_adp(u8 *ucode, int size)
{
	int i;
	u8 opcode_error, dummy;

	/* in case we fail */
	i34.adp_microcode_loaded = ADP_MICROCODE_NONE;

	dprintk("i34: ADP Microcode Version %d.%d\n", ucode[0], ucode[1]);
	if (! adp_write(ADP_BOOT) ) return 0;
	if (! adp_read(&opcode_error) ) return 0;
	if (opcode_error == ADP_OPCODE_ERROR) {
		dprintk("i34: i34_load_adp returns opcode error\n");
		return 0;
	}
	
	for (i=2; i<size; i++) {
		if (! adp_write(ucode[i]) ) return 0;
		if (! adp_read(&dummy) ) return 0;
	}

	adp_write(ADP_READ);
	adp_read(&opcode_error);
	if (opcode_error != ADP_NO_ERROR) {
		dprintk("i34: error loading ADP microcode\n");
		return 0;
	}

	dprintk("i34: ADP microcode loaded (%d bytes)\n", size);
	return 1;
}


int i34_adp_boot(int adp_microcode)
{
	int adp_loaded, size;
	u8 *ucode;

	if (i34.adp_microcode_loaded == adp_microcode)
	{
		dprintk("i34: ADP microcode not changed\n");
		return 1;
	}

	switch (adp_microcode)
	{
		case ADP_MICROCODE_AC3:
			ucode = adp_ac3;
			size  = sizeof(adp_ac3);
			break;
		
		case ADP_MICROCODE_AC36:
			ucode = adp_ac36;
			size  = sizeof(adp_ac36);
			break;

		case ADP_MICROCODE_LPCM:
			ucode = adp_lpcm;
			size  = sizeof(adp_lpcm);
			break;

		case ADP_MICROCODE_MPEG:
			ucode = adp_mpeg;
			size  = sizeof(adp_mpeg);
			break;

		case ADP_MICROCODE_PCM:
			ucode = adp_cdda;
			size  = sizeof(adp_cdda);
			break;

	default:
		dprintk("i34: invalid ADP microcode selection (%d)\n", adp_microcode);
		return 0;
	}

	adp_loaded = i34_load_adp(ucode, size);
	if (adp_loaded) i34.adp_microcode_loaded = adp_microcode;
	return adp_loaded;
}
	
int i34_load_dvp(u8 *ucode, int size)
{
	u8 buffer[60], temp;
	int i, count, limit;

	/* in case we fail */
	i34.dvp_microcode_loaded = DVP_MICROCODE_NONE;

	dprintk("i34: DVP Microcode Version %d.%d\n", ucode[0], ucode[1]);
	ucode += 2;
	limit = size;
	limit -= 2;
	/* prepare a chunk of microcode
	   swap bytes, because it's DVP microcode
	*/
	while (limit > 0)
	{
		for (i=0; (i<60) && (limit>0); i++, limit--)
		buffer[i] = ucode[i];
		count = i;
		for (i=1; i<count; i+=2)
		{
			temp = buffer[i-1];
			buffer[i-1] = buffer[i];
			buffer[i]   = temp;
		}
		ucode += count;
		if(! i34_write_parameter_seq(ZR_MICROCODE, (u16*)buffer, count>>1) )
		{
			dprintk("i34: error while downloading DVP microcode\n");
			return 0;
		}
	}
	dprintk("i34: DVP microcode loaded (%d bytes)\n", size); 

	/* zero-pad microcode */
	if (size < 16384)
	{
		count = (16384 - size);
		dprintk("i34: DVP microcode padded (%d bytes)\n", count);
		count >>= 2;
		for (i=0; i<count; i++)
		{
			i34_write_parameter(ZR_MICROCODE, 0);
		}
	}
	dprintk("i34: DVP microcode successfully loaded\n");
	return 1;
}

int i34_dvp_boot(int dvp_microcode)
{
	int dvp_loaded, size;
	u8 *ucode;

	if (i34.dvp_microcode_loaded == dvp_microcode)
	{
		dprintk("i34: DVP microcode not changed\n");
		return 1;
	}
	switch(dvp_microcode)
	{
		case DVP_MICROCODE_DVD:
			ucode = dvp_dvd;
			size = sizeof(dvp_dvd);
			break;

		case DVP_MICROCODE_VCD:
			ucode = dvp_vcd;
			size = sizeof(dvp_vcd);
			break;

		case DVP_MICROCODE_OTH:
			ucode = dvp_oth;
			size = sizeof(dvp_oth);
			break;

		default:
			dprintk("i34: invalid DVP microcode selection (%d)\n",
				dvp_microcode);
			return 0;
	}
	dvp_loaded = i34_load_dvp(ucode, size);
	if (dvp_loaded) i34.dvp_microcode_loaded = dvp_microcode;
	return dvp_loaded;
}


u8 init_file[] = {
#include "Zi34slt1.inc"
};

int i34_load_init_file(void)
{
	int i, count, size;
	u8 *codePtr;
	u8 buffer[64];
	
	if (i34.init_file_loaded) {
		dprintk("i34: init_file already loaded\n");
		return 1;
	}	
	size = sizeof(init_file) - 2;
	codePtr = &init_file[2];
	count = size;

	while(count>0) {
		for (i=1; (i<64) & (count>0); i+=2) {
			buffer[i]   = codePtr[i-1];
			buffer[i-1] = codePtr[i];
			count -= 2;
		}
		codePtr += i;
		if (! i34_write_parameter_seq(ZR_INITFILE_3, (u16*)buffer, i>>1) ) {
			dprintk("i34: error while loading init_file\n");
			return 0;
		}
	}
	dprintk("i34: SLOT MACHINE init_file loaded (%d bytes)\n", size);
	i34.init_file_loaded = 1;
	return 1;
}

/* rgb to yuv (YCrCb actually) by <inkling@nop.org>
   scales to 219 studio range (i hope) 4:4:4 value
*/
unsigned long rgb2yuv(long rgb24)
{
	int r, g, b;
	long iy, icb, icr;

	r = (rgb24 & 0xFF0000) >> 16;
	g = (rgb24 & 0xFF00) >> 8;
	b = (rgb24 & 0xFF);
                    
	/* you dont really wanna see how i got this formula */
	iy  =  16 + (( (16863 * r) + (32850 * g) +  (6351 * b)) >> 16);
	icr = 128 + (( (28689 * r) - (24090 * g) -  (4599 * b)) >> 16);
	icb = 128 + (( (-9636 * r) - (19053 * g) + (28689 * b)) >> 16);

       return (iy&255)<<16 | (icr&255)<< 8 | (icb&255);
}
                      
void i34_set_subpicture_palette(void)
{
	int i;
	long j;
       
	/* load subpicture (subtitle) palette with RGB values
	   bg is background (all black except 15), fg is foreground WRGBCMYK
	   [0] is white on black normal,  [15] is black on white inverse
	*/
	i34.SubPicturePalette[0]  = 0x000000; /* bg black	*/
	i34.SubPicturePalette[1]  = 0xFFFFFF; /* fg white	*/
	i34.SubPicturePalette[2]  = 0x000000; /* bg black	*/
	i34.SubPicturePalette[3]  = 0xFF0000; /* fg red		*/
	i34.SubPicturePalette[4]  = 0x000000; /* bg black	*/
	i34.SubPicturePalette[5]  = 0x00FF00; /* fg green	*/
	i34.SubPicturePalette[6]  = 0x000000; /* bg black	*/
	i34.SubPicturePalette[7]  = 0x0000FF; /* fg blue	*/
	i34.SubPicturePalette[8]  = 0x000000; /* bg black	*/
	i34.SubPicturePalette[9]  = 0x00FFFF; /* fg cyan	*/
	i34.SubPicturePalette[10] = 0x000000; /* bg black	*/
	i34.SubPicturePalette[11] = 0xFF00FF; /* fg magenta	*/
	i34.SubPicturePalette[12] = 0x000000; /* bg black	*/
	i34.SubPicturePalette[13] = 0xFFFF00; /* fg yellow	*/
	i34.SubPicturePalette[14] = 0xFFFFFF; /* bg white	*/
	i34.SubPicturePalette[15] = 0x000000; /* fg black	*/

	/* convert palette to YUV and swap 16 bit words (swapw not swapb)
	   NOTE: skip the swap step for big endian
	   this will become unnecessary once we read
	   the palette from the navpaks  <inkling@nop.org>
	*/

	for (i=0;i<16;i++) {
		/* convert from easier to read RGB24 to YUV219 */
		j = rgb2yuv(i34.SubPicturePalette[i]);

		/* NOTE: skip this swap step for big endian */
		j = (j & 0xffff)<< 16 | (j & 0xffff0000)>>16;
		/* swap just the words, not the entire byte order */

		/* dprintk("i34: palette[%x] = %08X \n",i,j); */
		i34.SubPicturePalette[i] = j;
	}
}

/* Implement the protocol to read/write to/from the DVP data buffer
   (see p.106 and following). These functions always process 64 bytes
   at a time (32 words)
*/
#define DVPIBE	0x0080
#define DVPVAL	0x0100
#define DVPOBF	0x4000
int i34_read_dvp_data(u16 *data)
{
	u16 status1;
	u16 isr;
	int i;

	/* FIXME: this should go into the interrupt routine */
	isr = i34_read(ZR_ISR);
	if (isr & DVPOBF) {
		for (i=0; i<32; i++)
			data[i] = i34_read(ZR_DVP_DATA);
		status1 = i34_read(ZR_STATUS1);
		if (status1 & DVPVAL) {
			dprintk("i34: data read from DVP has been corrupted\n");
			return 0;
		}
		return 1;
	}
	return 0;
}

int i34_write_dvp_data(u16 *data)
{
	u16 status1;

	status1 = i34_read(ZR_STATUS1);
	if (status1 & DVPIBE) {
		return 1;
	}
	dprintk("i34: DVP not ready to receive data\n");
	return 0;
}

int i34_read_global_delta(void)
{
	u16 dvp_buffer[32];

	i34_read_dvp_data(dvp_buffer);
	/* see p. 147 */
	if (dvp_buffer[0] == 0x4105) {
		dprintk("i34: GLOBAL_DELTA = %dl\n", (dvp_buffer[1]<<16) + dvp_buffer[2]);
	}
	return 0;
}


/* Prepares the display
   while state = init_pclk (see p. 80)
*/
int i34_init_video_port(void)
{
	int cstate;

	/* Video configuration */
	i34.VideoStandard = VIDEO_MODE_NTSC;
	/* i34.vs = &ntsc1; */
	i34.cbs.CBHStart = 0x0A;
	i34.cbs.CBVStart = 0;
	i34.cbs.CBHEnd = 0;
	i34.cbs.CBVEnd = 0;
	i34_set_subpicture_palette();
	i34.ckey.key[0] = 0x10;
	i34.ckey.key[1] = 0x80;
	i34.ckey.key[2] = 0x80;

	i34.AutoScaling = 3;
	i34.AutoPanScan = 1;
	i34.ScaleRatio = 0x1111;

	i34.BgSwitch = 1;
	i34.VidEntry = 0;
	i34.LastPic = 0x200;
	i34.VidFloat = 0; /* 0x40; */

	i34.SpSwitch = 5;
	i34.CaptionSwitch = 0;

	i34.video_sid = 0x0000;	/* first video stream ID encountered  (see p. 128) */
	i34.video_tolerance = 0x1000;
	i34.video_sync_mode = 1;	/* video DOES synchronize to SCLK (p. 142) */

	dprintk("i34: i34_init_video_port\n");

	i34_video_standard(i34.VideoStandard);
	i34_composite_blank_sync(&i34.cbs);
	i34_color_key(&i34.ckey);
	i34_write_parameter(ZR_CAPTIONOFFSET, 21);
	i34_write_parameter(ZR_BACKGROUNDSWITCH, 1);
	i34_write_parameter(ZR_STARTDISPLAY, 1);

	if (! i34_state(&cstate) ) return 0;

	/* we should now be in state = init_display */
	if (cstate != 0x03) {
		dprintk("i34: warning device not in 'init_display' state after init_video_port\n");
		return 0;
	}
	return 1;
}

int i34_load_palette(u32 *palette)
{
	return i34_write_parameter_seq(ZR_SPPALETTE, (u16*)palette, 32);
}

int i34_change_audio_sid(int sid)
{
	dprintk("i34: i34_change_audio_SID\n");
	adp_simple_command(ADP_STOPF);
	/* see p. 258 */
	i34_write_parameter(ZR_AUDSID, 0x00FF);
	i34_wait(500);
	/* FIXME: if audioStandard like AC3 or PCM changes
	     ADP_CFG
	     ADP_SPDIFCS
	     ADP_audioStandardCFG
	*/
	adp_audio_sync(i34.audio_sync_mode);
	adp_simple_command(ADP_PLAY);
	i34.audio_sid = sid;
	i34_write_parameter(ZR_AUDSID, sid);
	adp_simple_command(ADP_UNMUTE);
	return 1;
}

int i34_init_audio_port(void)
{
	dprintk("i34: i34_init_audio_port\n");

	/* Audio Configuration */
	i34.audio_tolerance = 0x1000;
	i34.audio_sync_mode = 3; 				/* clock master then audio master (p. 142 & 221) */
	i34.OutputSamplingRate = 0; 			/* 48KHz (p. 229) */
	i34.audio_gain = 0x7FFFF;			/* max by default */

	/* This configures the serial interface between the decoder
	   and the external DAC.
	   see p. 226 ZR HB
	   see p. 10 AD1857 HB
	 */
	i34.serial.cpa = i34.serial.cpb = 0;	/* negative edge */
	i34.serial.err = 0;			/* error output pin disabled */
	i34.serial.spo = 1;			/* S/PDIF acts as a S/PDIF transmiter */
	i34.serial.fra = i34.serial.frb  = 1;	/* 32 bits frame size */
	i34.serial.inw = i34.serial.outw = 0;	/* 20 bits input/output word */
	/* ALRCLK low for left channel, left justified + 1 bit (ABCLK cycle) */
	i34.serial.isp = i34.serial.osp = 0;
	i34.serial.fmb = i34.serial.fma = 1;

	adp_clean();
	adp_rom_version();

	i34.amclk = 256; /* dbarth: inkling patch said 192, but the external DAC is not
			    supposed to accept such a rate: tell me which one is correct */
	adp_plltab();

	i34.adp_microcode_loaded = ADP_MICROCODE_NONE;

	return 1;
}

int i34_configure_audio(void)
{
	dprintk("i34: i34_configure_audio\n");

	/* set the 2 GPAIO lines in output mode to drive the external DAC and multiplexer
	   Thanks to Gernot Ziegler <gz@lysator.liu.se> for this information :
	   GPA I/O 0 leads to Pin 10, Philips HT4052 (acting as select pin S0 ?)
	   GPA I/O 1 leads to Pin 2, AD 1857 (!PD/!RST, make HI for normal operation, says the manual) 
	*/
	adp_config_gpaio(1, 1);
	adp_config_gpaio(2, 2);
	adp_set_gpaio(2, 0);	/* reset the DAC */

	/* PLLCFG must be done before CFG command p. 229 (thanks inkling@nop.org) */
	adp_pllcfg(i34.OutputSamplingRate);

	adp_clean();

	/* the CFG command must be followed by an AC3 or sth (p. 227) */
	/* see also table 47 p. 114 to check if it's valid to send a command at a certain state */
	adp_config_serial(&i34.serial);
	adp_config_spdif();

	adp_config_gpaio(1, 1);
	adp_set_gpaio(1, 0);	/* select S0 on the analog demultiplexer */
	adp_config_gpaio(2, 2);
	adp_set_gpaio(2, 2);	/* boot up the external DAC */

	/* FIXME: usefull ?
	   Yes perhaps, now that the DAC is initialized lately to avoid dirty clicks
	   while we scan for available devices (see vobcat.c)
	*/
	i34_wait(1000);

	adp_set_function();

	i34_write_parameter(ZR_AUDSID, i34.audio_sid);

	adp_audio_tolerance(i34.audio_tolerance);
	adp_audio_sync(i34.audio_sync_mode);
	i34_write_parameter(ZR_AUDPORTDELAY, 0);

	adp_simple_command(ADP_PLAY);
	adp_simple_command(ADP_UNMUTE);
	adp_audio_gain(i34.audio_gain);

	return 1;
}

int i34_change_video_sid(int sid)
{
	i34.video_sid = sid;
	i34_write_parameter(ZR_VIDSID, i34.video_sid);
	return 1;
}

int i34_change_subpic_sid(int sid)
{
	i34.subpic_sid = sid;
	i34_write_parameter(ZR_SPSID, i34.subpic_sid);
	return 1;
}

int i34_configure_display(void)
{
	/* Scaling & Panning (p. 158) */
	i34_write_parameter(ZR_AUTOPANSCAN, i34.AutoPanScan);
	i34_write_parameter(ZR_PANSCANBASEX, 0);
	i34_write_parameter(ZR_PANSCANBASEY, 0);
	i34_write_parameter(ZR_PANSCANOFFSETX, 0);
	i34_write_parameter(ZR_PANSCANOFFSETY, 0);
	i34_write_parameter(ZR_AUTOSCALING, i34.AutoScaling);
	i34_write_parameter(ZR_SCALERATIO, i34.ScaleRatio);
	if (!i34.AutoPanScan) {
		i34_write_parameter(ZR_IMAGESTARTX, i34.vs->ActiveStartX);
		i34_write_parameter(ZR_IMAGESTARTY, i34.vs->ActiveStartY);
		i34_write_parameter(ZR_IMAGEENDX, i34.vs->ActiveEndX);
		i34_write_parameter(ZR_IMAGEENDY, i34.vs->ActiveEndY);
		i34_write_parameter(ZR_IMAGESIZEX, i34.vs->ActiveSizeX);
		i34_write_parameter(ZR_IMAGESIZEY, i34.vs->ActiveSizeY);
	}
	i34_write_parameter(ZR_DVPGEN3, 0);
	i34_write_parameter(ZR_DISCTYPE, 0);
	i34_change_parameter(ZR_BITSTREAMSELECT, i34.VidEntry, 1);
	i34_change_parameter(ZR_PLAYBACKMODE, i34.LastPic, 0x200);
	i34_change_parameter(ZR_PLAYBACKMODE, i34.VidFloat, 0x40);
	i34_change_parameter(ZR_PLAYBACKMODE, 0x10, 0x10);	/* automagically sets the background to black while decoding */
	i34_load_palette(i34.SubPicturePalette);
	i34_change_parameter(ZR_SPSWITCH, i34.SpSwitch, 7);
	i34_write_parameter(ZR_CAPTIONSWITCH, i34.CaptionSwitch);
	i34_write_parameter(ZR_OSDSWITCH, 0);

	i34_change_parameter(ZR_DVPGEN2, 0x80, 0x80);	/* error concealment : less freezes, more greenblocks */
	i34_change_parameter(ZR_DVPGEN2, 1, 0);		/* overwrite NV_PCK (see p. 118) */

	i34_write_parameter(ZR_VIDSID, i34.video_sid);
	i34_write_parameter(ZR_SPSID, 0x0020); 		/* sub-picture stream encountered */

	i34_write_parameter(ZR_STARTDISPLAY, 1);	/* validate the new parameters (still usefull ?) */
	
	i34_write_parameter(ZR_VIDTOLERANCE, i34.video_tolerance);
	i34_write_parameter(ZR_VIDSYNCMODE, i34.video_sync_mode);
	i34_write_parameter(ZR_VIDPORTDELAY, 0);

	return 1;
}

int i34_select_bitstream(int bs)
{
	u16 cbs;

	cbs = bs & 0x1F;	/* sanity limit bitstream value */
	
	printk("i34: select_bitstream %02X\n",cbs);
	switch (cbs)
	{
		/* make sure we only select a valid bitstream type.
		   do all these break's need to be in here? i don't know.
		   let me go re-read my k&r c book.
		   -inkling */

		case BITSTREAM_VCD:
			break;
		case BITSTREAM_DVD_VOB:
			/* FIXME: default to loading DVD & AC3 microcode
			   these two REALLY belong in the user app using
			   the new ioctls that have been made for them
			   because only the user app knows what it needs.
			   -inkling */
			i34_dvp_boot(DVP_MICROCODE_DVD);
			i34_adp_boot(ADP_MICROCODE_AC3);
			break;
		case BITSTREAM_CDDA_PCM:
			break;
		case BITSTREAM_VCD_AUX:
			break;
		case BITSTREAM_DVD_NAV:
			break;
		case BITSTREAM_MPEG12_PS:
			break;
		case BITSTREAM_MPEG12_V_ES:
			break;
		case BITSTREAM_MPEG12_AV_PES:
			break;
		case BITSTREAM_MPEG12_A_ES:
			break;
		case BITSTREAM_AC3_A_ES:
			break;
		case BITSTREAM_LPCM_A_ES_DVD:
			break;
		default:
			dprintk("i34: invalid bitstream selection (0x%02X)\n",
				cbs);
			return 0;
	}

	i34_change_parameter(ZR_BITSTREAMSELECT, cbs << 11, 0xF800);
	dprintk("i34: codec bitstream 0x%02X selected (0x%04X)\n",cbs,cbs<<11);
	i34_load_init_file();
	return 1;
}

int i34_enable_interrupts(int interrupts)
{
	i34_write(ZR_IMR, interrupts);
	return 0;
}

int i34_init(void)
{
	int i;

	/* According to Annex C p. 248 of the ZR36710 Handbook */
	
	dprintk("i34: i34_init\n");

	spin_lock_init(i34.lock);

	/* defaults for parameter registers */
	for (i=0; i<0xff; i++)
		i34.shadow_register[i] = 0;

	i34.CodeBurstLen = 64;
	i34.init_file_loaded = 0;

	/* Stage 1 : Reset */
//	i34_reset(0);
//	i34_reset(0);

	/* State = init_pclk */

	/* NumDRAMDev = 0; CodeSource = 1; Decryption bypass */
	i34_write_parameter(ZR_SYSCONFIG, (0<<13) | (1<<12) | (i34.CodeBurstLen<<4) | (0<<3));

	i34_change_parameter(0xA0, 0x80, 0x80);
	i34_change_parameter(0xA0, 0x00, 0x80);
	i34_change_parameter(0xA0, 0x0F, 0x0F);
	i34_change_parameter(0xA4, 0x00, 0x40);
	i34_change_parameter(0xA4, 0x01, 0x07);
	
	i34_write_parameter(ZR_SDCONFIG, 0);

	/* interrupt on NAVREADY */
	i34_write(ZR_IMR, 0x7FFF); 

	/* Audio Port */
	if (! i34_init_audio_port() ) return 0;
	
	/* Video Port */
	if (! i34_init_video_port() ) return 0;

	/* Bitsream, Disc and Entry Point selection */
	/* i34_select_bitstream(BITSTREAM_MPEG_PES); */
	i34_select_bitstream(BITSTREAM_DVD_VOB);

	i34_configure_display();

	/* force the device into the idle (0x30) state */
	i34_send_command(I34_CMD_START, 0);
	i34_send_command(I34_CMD_END, 0);

	return 1;
}

int i34_start(void)
{
	if (! i34_configure_audio() ) return 0;
	if (! i34_send_command(I34_CMD_START, 0) ) return 0;
	/* FIXME: hack to kick the ADP out of "hang" mode */
	if (! adp_decoder_status() ) return 0;
	/* Danger Will Robinson ! */
	/* i34_change_parameter(ZR_VIDSYNCMODE, 2, 2); */
	return 1;
}

int i34_stop(void)
{
	adp_simple_command(ADP_STOPF);
	if (!i34_send_command(I34_CMD_END, 1)) return 0;
	/* FIXME: send SEQEND ? */
	/* i34_send_command(I34_CMD_SEQEND, 0); */

	/* "release" the DAC : without this your CD player will stay muted... */
	adp_config_gpaio(3, 0);

	return 1;
}

int i34_pause(void)
{
	if (!i34_send_command(I34_CMD_PAUSE, 0))
		return -EINVAL;
	return 0;
}

int i34_continue(void)
{
	if (!i34_send_command(I34_CMD_CONT, 0))
		return -EINVAL;
	return 0;
}

int i34_status(struct mgadvd_status_s *st)
{
	int i;

	st->status0 = i34_read(ZR_STATUS0);
	st->status1 = i34_read(ZR_STATUS1);
	st->status2 = i34_read(ZR_STATUS2);
	st->sclk = i34_read_sclk();
	st->video_buffer_status = i34_read(ZR_VIDEOBUF);
	st->audio_buffer_status = i34_read(ZR_AUDIOBUF);
	if (!adp_decoder_status()) return -EBUSY;
	for (i=0; i<16; i++)
		st->adp_status_reply[i] = adp_buffer[i];

	return 0;
}
