/*
 *
 * mga_vid_test.c
 *
 * Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 * Sept 1999
 *
 * This software has been released under the terms of the GNU Public
 * license. See http://www.gnu.org/copyleft/gpl.html for details.
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>

#include <linux/videodev.h>

#ifndef VIDEO_WINDOW_CHROMAKEY
#define VIDEO_WINDOW_CHROMAKEY	16
#endif /* VIDEO_WINDOW_CHROMAKEY */

#ifndef AARONS_TYPES
typedef unsigned long uint_32;
typedef unsigned char uint_8;
#endif                                                                          

struct video_window vidwin;
struct video_picture picfmt;
struct video_buffer vidbuf;

uint_8 *mga_vid_base;

uint_8 *y_image;
uint_8 *cr_image;
uint_8 *cb_image;

void allocate_video_buffers(int sw, int sh)
{
	y_image = (uint_8 *)malloc(sw * sh);
	cr_image = (uint_8 *)malloc(sw * sh);
	cb_image = (uint_8 *)malloc(sw * sh);
}

void lookup_videodev(char **result)
{
	int i, fd;
	char devname[32];
	struct video_capability capability;

	*result = NULL;

	printf("scanning video devices...");
	for (i=0; i<10; i++) {
		sprintf(devname, "/dev/video%d", i);
		printf("\n%s...", devname);
		fd = open(devname, O_RDONLY);
		if (fd == -1)
			continue;
		if (ioctl(fd, VIDIOCGCAP, &capability) == 0) {
			printf("%s\n", capability.name);
			/* is it an OVERLAY only device ? */
			if ((capability.type & VID_TYPE_OVERLAY) &&
			    !(capability.type & VID_TYPE_CAPTURE)) {
				/* we've found a capable device */
				*result = strdup(devname);
				close(fd);
				return;
			}
		}
		close(fd);	
	}
	fprintf(stderr, "unable to find an overlay device\n");
	exit(2);
}

void
write_frame_YUV420(uint_8 *y,uint_8 *cr, uint_8 *cb)
{
	uint_8 *dest;
	uint_32 bespitch,h,w;

	dest = mga_vid_base;
	bespitch = vidbuf.bytesperline;

	for(h=0; h < vidbuf.height; h++) 
	{
		memcpy(dest, y, vidbuf.width);
		y += vidbuf.width;
		dest += bespitch;
	}

	for(h=0; h < vidbuf.height/2; h++) 
	{
		for(w=0; w < vidbuf.width/2; w++) 
		{
			*dest++ = *cb++;
			*dest++ = *cr++;
		}
		dest += bespitch - vidbuf.width;
	}
}

void
write_frame_YUV420P(uint_8 *y,uint_8 *cr, uint_8 *cb)
{
	uint_8 *dest;
	uint_32 bespitch,h;

	dest = mga_vid_base;
	bespitch = vidbuf.bytesperline;

	for(h=0; h < vidbuf.height; h++) 
	{
		memcpy(dest, y, vidbuf.width);
		y += vidbuf.width;
		dest += bespitch;
	}

	for(h=0; h < vidbuf.height/2; h++) 
	{
		memcpy(dest, cb, vidbuf.width/2);
		cb += vidbuf.width/2;
		dest += bespitch/2;
	}

	for(h=0; h < vidbuf.height/2; h++) 
	{
		memcpy(dest, cr, vidbuf.width/2);
		cr += vidbuf.width/2;
		dest += bespitch/2;
	}
}

void write_frame(uint_8 *y,uint_8 *cr, uint_8 *cb)
{
	switch (picfmt.palette) {
	case VIDEO_PALETTE_YUV420:
		write_frame_YUV420(y,cr,cb);
		break;
	case VIDEO_PALETTE_YUV420P:
		write_frame_YUV420P(y,cr,cb);
		break;
	default:
		fprintf(stderr, "unsupported palette format\n");
	}
}

void
draw_cool_pattern(void)
{
	int i,x,y;

	i = 0;
	for (y=0; y<vidbuf.height; y++) {
		for (x=0; x<vidbuf.width; x++) {
			y_image[i++] = x*x/2 + y*y/2 - 128;
		}
	}

	i = 0;
	for (y=0; y<vidbuf.height/2; y++) 
		for (x=0; x<vidbuf.width/2; x++) 
		{
				cr_image[i++] = x - 128;
		}

	i = 0;
	for (y=0; y<vidbuf.height/2; y++) 
		for (x=0; x<vidbuf.width/2; x++) 
		{
				cb_image[i++] = y - 128;
		}
}

void
draw_color_blend(void)
{
	int i,x,y;

	i = 0;
	for (y=0; y<vidbuf.height; y++) {
		for (x=0; x<vidbuf.width; x++) {
			y_image[i++] = 0;
		}
	}

	i = 0;
	for (y=0; y<vidbuf.height/2; y++) 
		for (x=0; x<vidbuf.width/2; x++) 
		{
				cr_image[i++] = x - 128;
		}

	i = 0;
	for (y=0; y<vidbuf.height/2; y++) 
		for (x=0; x<vidbuf.width/2; x++) 
		{
				cb_image[i++] = y - 128;
		}
}

void
show_maps(void)
{
	char buf[256];
	int fd;

	fd = open("/proc/self/maps", O_RDONLY);

	if (fd == -1) {
		fprintf(stderr, "unable to access memory maps\n");
		return;
	}

	fputs("** Memory mappings **\n", stderr);
	while (read(fd, buf, 256)) {
		fputs(buf, stderr);
	}
	fputs("\n", stderr);
}

int 
main(int argc, char *argv[])
{
	struct video_channel vchan;
	unsigned long bufsize;
	char *devname;
	int f;

	lookup_videodev(&devname);

	f = open(devname, O_RDWR);

	if(f == -1)
	{
		fprintf(stderr,"Couldn't open driver\n");
		exit(1);
	}

	vchan.channel = 0;
	if (ioctl(f, VIDIOCGCHAN, &vchan))
		perror("VIDIOCGCHAN");

	vchan.norm = VIDEO_MODE_PAL;
	if (ioctl(f, VIDIOCSCHAN, &vchan))
		perror("VIDIOCSCHAN");

	vidwin.x = 0;
	vidwin.y = 0;
	vidwin.width = 768;
	vidwin.height = 576;
	vidwin.flags = VIDEO_WINDOW_CHROMAKEY;
	
	if (ioctl(f, VIDIOCSWIN, &vidwin))
		perror("VIDIOCSWIN");

	if (ioctl(f, VIDIOCGPICT, &picfmt))
		perror("VIDIOCGPICT");

	if (ioctl(f, VIDIOCGFBUF, &vidbuf))
		perror("VIDIOCGFBUF");

	allocate_video_buffers(vidbuf.width, vidbuf.height);

	bufsize = vidbuf.bytesperline * vidbuf.height * 2; /* YUV420 */
	mga_vid_base = (uint_8*)mmap(0, bufsize, PROT_WRITE, MAP_SHARED, f, 0);

	if (ioctl(f, VIDIOCCAPTURE, 1))
		perror("VIDIOCCAPTURE");

	write_frame(y_image,cr_image,cb_image);
	printf("(1) There should be a green square, offset by 10 pixels from\n"
			   "    the upper left corner displayed\n");
	sleep(3);

	draw_cool_pattern();
	write_frame(y_image,cr_image,cb_image);
	printf("(2) There should be a cool mosaic like pattern now.\n");
	sleep(3);

	draw_color_blend();
	write_frame(y_image,cr_image,cb_image);
	printf("(3) There should be a color blend with black, red, purple, blue\n"
			   "    corners (starting top left going CW)\n");
	sleep(3);

	close(f);
	return 0;
}
