/****************************************************************************
 * back-end scaler
 * (based on code from Benoit Mathieu and Aaron Holtzman)
 ****************************************************************************/

#define __NO_VERSION__
#include <linux/module.h>
#include <linux/videodev.h>	/* for VIDEO_PALETTE_... */
#include "mgacore.h"
#include "mgabes.h"

EXPORT_SYMBOL(mgabes_init);
EXPORT_SYMBOL(mgabes_set_source);
EXPORT_SYMBOL(mgabes_set_window);
EXPORT_SYMBOL(mgabes_set_frame);
EXPORT_SYMBOL(mgabes_set_mode);
EXPORT_SYMBOL(mgabes_set_colorkey);
EXPORT_SYMBOL(mgabes_activate);
EXPORT_SYMBOL(mgabes_desactivate);
EXPORT_SYMBOL(mgabes_pitch_rounding);

/* Backend scaler registers */
#define BESA1ORG        0x3d00
#define BESA2ORG        0x3d04
#define BESB1ORG        0x3d08
#define BESB2ORG        0x3d0c
#define BESA1CORG       0x3d10
#define BESA2CORG       0x3d14
#define BESB1CORG       0x3d18
#define BESB2CORG       0x3d1c
#define BESCTL          0x3d20
#define BESPITCH        0x3d24
#define BESHCOORD       0x3d28
#define BESVCOORD       0x3d2c
#define BESHISCAL       0x3d30
#define BESVISCAL       0x3d34
#define BESHSRCST       0x3d38
#define BESHSRCEND      0x3d3c
#define BESLUMACTL	0x3d40	/* G400 */
#define BESV1WGHT       0x3d48
#define BESV2WGHT       0x3d4c
#define BESHSRCLST      0x3d50
#define BESV1SRCLST     0x3d54
#define BESV2SRCLST     0x3d58
#define BESA1C3ORG	0x3d60	/* G400 */
#define BESA2C3ORG	0x3d64	/* G400 */
#define BESB1C3ORG	0x3d68	/* G400 */
#define BESB2C3ORG	0x3d6c	/* G400 */
#define BESGLOBCTL      0x3dc0
#define BESSTATUS       0x3dc4

#define VCOUNT		0x1e20

#define XMULCTRL      0x19
#define BPP_8         0x00
#define BPP_15        0x01
#define BPP_16        0x02
#define BPP_24        0x03
#define BPP_32_DIR    0x04
#define BPP_32_PAL    0x07

/*
 * These registers closely match the names of the hardware spec :
 * (check p. 3-59 (99) and following in the G400 manual)
 */
struct bes_registers_s
{
	u32 besctl;
	u32 besglobctl;
	u32 beslumactl;
	u32 bespitch;
	u32 besa1c3org;
	u32 besa1corg;
	u32 besa1org;
	u32 besa2c3org;
	u32 besa2corg;
	u32 besa2org;
	u32 besb1c3org;
	u32 besb1corg;
	u32 besb1org;
	u32 besb2c3org;
	u32 besb2corg;
	u32 besb2org;
	u32 beshcoord;
	u32 beshiscal;
	u32 beshsrcst;
	u32 beshsrcend;
	u32 beshsrclst;
	u32 besvcoord;
	u32 besviscal;
	u32 besv1srclst;
	u32 besv1wght;
	u32 besv2srclst;
	u32 besv2wght;
};

struct mga_bes_s
{
	/*
	 * mode (interlaced, double buffered) and options like
	 * accelerated 2x zoom, hfilter, ...
	 */
	u32 mode;

	/* source video */
	int srcheight, srcwidth;
	int palette;
	u32 buf_a, buf_b;

	/* frame to display */
	int frame;

	/* overlay video window */
	int x, y;
	int dstwidth, dstheight;
	u32 chromakey;

	/* registers */
	struct bes_registers_s regs;

	/* host card */
	struct mga_dev_s *mga;
};

/* FIXME: this is a hack */
static struct mga_bes_s unique_bes;


/**
 *	mgabes_init - 	initialize the BES
 *	@mga:	mga instance
 *
 *	This function initializes the BackEnd Scaler, responsible for
 *	displaying a scaled video overlay.
 */
void mgabes_init(struct mga_dev_s *mga)
{
	struct mga_bes_s *bes;

	if (!mga) return;

	bes = &unique_bes;

	mga->bes = bes;
	bes->mga = mga;

	/* default settings */
	bes->x = 0;
	bes->y = 0;
	bes->srcwidth = 720;
	bes->srcheight = 480;
	bes->dstwidth = 720;
	bes->dstheight = 480;
	bes->chromakey = 0;
	bes->palette = VIDEO_PALETTE_YUV422;
	bes->buf_a = 0;
	bes->buf_b = 0;
	bes->frame = BES_FRAME_A;
	bes->mode = 0;
}

/**
 *	mgabes_pitch_rounding - 	return the pitch for a given width
 *	@bes:		bes instance
 *	@width:		width of the video frame
 *
 *	This function returns the pitch value depending on the video
 *	format, the video buffer width and the card type.
 *
 *	FIXME:	This could be a macro/inline function usable from
 *	userland.
 */
unsigned int mgabes_pitch_rounding(struct mga_bes_s *bes, int width)
{
	unsigned int pitch;

	/*
	 * see p. 4-116 (566) G400 spec
	 * and p. 4-99  (477) G200 spec
	 */
	if ((bes->palette == VIDEO_PALETTE_YUV420) ||
	    (bes->palette == VIDEO_PALETTE_YUV420P)) {
		if (bes->mode & BES_IS_G400)
			pitch = (width + 7) & ~7;
		else
			pitch = (width + 3) & ~3;
	} else { /* palette == VIDEO_PALETTE_YUV422 */
		if (bes->mode & BES_IS_G400)
			pitch = (width + 31) & ~31;
		else
			pitch = (width + 7) & ~7;
	}
	return pitch;
}

static void compute_horizontal_scaling(struct mga_bes_s *bes)
{
	u32 intrep, ifactor, ifactorbetter, condition, roundoff,
		acczoom; 
	u32 offsetleft, offsetright, cropleft, cropright;
	u32 cropped_width;
	struct bes_registers_s *regs = &bes->regs;

	/* we do not crop the source window */
	cropleft = 0;
	cropright = 0;
	/* we do not crop the destination window either */
	offsetleft = 0;
	offsetright = 0;

	/* see section 4.10.2 p. 4-118 (566) in the G400 spec */ 
	regs->beshsrclst = bes->srcwidth - 1;

	/* 4.10.2.1 : Horizontal Inverse Scaling Factor */
	cropped_width = bes->srcwidth - cropleft - cropright;
	if (bes->mode & BES_HFILTER)
		intrep = (bes->dstwidth == cropped_width) ? 0 : 1;
	else
		intrep = (bes->dstwidth >= cropped_width) ? 0 : 1;
	ifactor = ((cropped_width - intrep) << 14) / (bes->dstwidth - intrep);
	ifactorbetter = ((cropped_width - intrep) << 20) / (bes->dstwidth - intrep) + intrep;
	condition = ( ((ifactorbetter * (bes->dstwidth-1)) >> 20) > ((ifactor * (bes->dstwidth-1)) >> 14) );
	if (condition)
		roundoff = !(bes->mode & BES_HFILTER) ? 1 : 0;
	else
		roundoff = 0;
	acczoom = (bes->mode & BES_ACC2XZOOM) ? 2 : 1;
	regs->beshiscal = (acczoom * ifactor) + roundoff;
	if (regs->beshiscal >= (32 << 14))
		regs->beshiscal = (32 << 14) - 1;

	/* Horizontal source positioning */
	regs->beshsrcst = (cropleft << 14) + offsetleft * (ifactor + roundoff);
	regs->beshsrcend = regs->beshsrcst + ((bes->dstwidth - offsetleft - offsetright - 1) / acczoom)
	             * (acczoom * ifactor + roundoff);

	/* adjust some registers */
	regs->beshsrcend <<= 2;
	regs->beshsrcst <<= 2;
	regs->beshiscal <<= 2;
}

static void
compute_vertical_source_positionning(struct mga_bes_s *bes, int field, u32 sc)
{
	struct bes_registers_s *regs = &bes->regs;
	u32 offsettop, offsetbottom;
	u32 croptop, cropbottom;
	u32 bes420pl, dataformat, format420, bes3plane;
	u32 weight, vsrclst, vsrcstp, val, absval, tmp;
	u32 base_address, origin_address;
	u32 chroma_plane_base_address, chroma_plane_origin_address;
	u32 chroma_3plane_base_address, chroma_3plane_origin_address;

	/* we do not crop the source window */
	croptop = 0;
	cropbottom = 0;
	/* we do not crop the destination window either */
	offsettop = 0;
	offsetbottom = 0;

	/*
	 * Vertical source positioning
	 * (see p. 4-124 (572) in the G400 spec)
	 *
	 */
	switch (bes->palette) {
	case VIDEO_PALETTE_YUV420:
		bes420pl = 1;
		bes3plane = 0;
		dataformat = 1;
		format420 = 1;
		break;
	case VIDEO_PALETTE_YUV420P:
		bes420pl = 1;
		bes3plane = 1;
		dataformat = 1;
		format420 = 2;
		break;
	case VIDEO_PALETTE_YUV422:
	default:
		bes420pl = 0;
		bes3plane = 0;
		dataformat = 2;
		format420 = 1;
		break;
	}

	if (regs->besviscal >= (1 << 14)) {
		/* If downscaling */
		val = croptop + ((offsettop * regs->besviscal + sc) >> 14); 
		origin_address = val * regs->bespitch * dataformat;

		chroma_plane_origin_address = (val / (2 * format420)) * regs->bespitch;

		chroma_3plane_origin_address = (val / 4) * regs->bespitch;

		if ((bes->mode & BES_INTERLACED) /* should be BES_DEINTERLACE */
		    && ((field == BES_FIELD_A2) || (field == BES_FIELD_B2)))
			if (offsettop == 0)
				weight = -((1<<13) + (regs->besviscal & 0x3FFF)); 
			else
				weight = -(1<<13) + offsettop * regs->besviscal;
		else
			weight = offsettop * regs->besviscal;

		vsrclst = bes->srcheight - 1 - val;
		vsrcstp = val;
	} else {
		/* FIXME: sc>>14 ? */
		absval = croptop + ((abs((signed int)offsettop - (sc>>14)) * regs->besviscal) >> 14);
		origin_address = absval * regs->bespitch * dataformat;

		chroma_plane_origin_address = (absval / (2 * format420)) * regs->bespitch;
		chroma_3plane_origin_address = (absval / 4) * regs->bespitch;

		if ((bes->mode & BES_INTERLACED) /* should be BES_DEINTERLACE */
		    && ((field == BES_FIELD_A2) || (field == BES_FIELD_B2)))
			if ((offsettop * regs->besviscal) >= (1<<13))
				weight = -(1<<13) + offsettop;
			else
				weight = -((1<<13) + offsettop * regs->besviscal);
		else
			weight = offsettop * regs->besviscal;

		vsrclst = bes->srcheight - 1 - absval;
		vsrcstp = absval;
	}
	/* FIXME: add mirror parameters if mirroring is supported */

	tmp = (weight << 2) | (((weight & (1<<31)) ? 1 : 0) << 16);

	switch (field) {
	default:
	case BES_FIELD_A1:
	case BES_FIELD_B1:
		regs->besv1wght = tmp;
		regs->besv1srclst = vsrclst;
		regs->besctl = (vsrcstp & (1<<14)) ? (1<<6) : 0;
		break;
	case BES_FIELD_A2:
	case BES_FIELD_B2:
		regs->besv2wght = tmp;
		regs->besv2srclst = vsrclst;
		regs->besctl = (vsrcstp & (1<<14)) ? (1<<7) : 0;
		break;
		break;
	}
	regs->besctl |= (bes420pl << 17);

	chroma_plane_base_address = regs->bespitch * dataformat * bes->srcheight;
	chroma_3plane_base_address = (regs->bespitch * bes->srcheight) / 4;

	switch (field) {
	default:
	case BES_FIELD_A1:
		base_address = bes->buf_a;
		regs->besa1org = base_address + origin_address;
		regs->besa1corg = base_address + chroma_plane_base_address +
			chroma_plane_origin_address;
		regs->besa1c3org = base_address + chroma_3plane_base_address +
			chroma_3plane_origin_address;
		break;
	case BES_FIELD_A2:
		base_address = bes->buf_a;
		regs->besa2org = base_address + origin_address;
		regs->besa2corg = base_address + chroma_plane_base_address +
			chroma_plane_origin_address;
		regs->besa2c3org = base_address + chroma_3plane_base_address +
			chroma_3plane_origin_address;
		break;
	case BES_FIELD_B1:
		base_address = bes->buf_b;
		regs->besb1org = base_address + origin_address;
		regs->besb1corg = base_address + chroma_plane_base_address +
			chroma_plane_origin_address;
		regs->besb1c3org = base_address + chroma_3plane_base_address +
			chroma_3plane_origin_address;
		break;
	case BES_FIELD_B2:
		base_address = bes->buf_b;
		regs->besb2org = base_address + origin_address;
		regs->besb2corg = base_address + chroma_plane_base_address +
			chroma_plane_origin_address;
		regs->besb2c3org = base_address + chroma_3plane_base_address +
			chroma_3plane_origin_address;
		break;
	}
}

static void
compute_vertical_scaling(struct mga_bes_s *bes)
{
	struct bes_registers_s *regs = &bes->regs;
	u32 offsettop, offsetbottom;
	u32 croptop, cropbottom, cropped_height;
	u32 intrep, ifactor, ifactorbetter, condition, roundoff;
	u32 subpixelcomp;

	/* we do not crop the source window */
	croptop = 0;
	cropbottom = 0;
	/* we do not crop the destination window either */
	offsettop = 0;
	offsetbottom = 0;

	cropped_height = bes->srcheight - croptop - cropbottom;
	
	/* Inverse Vertical scaling factor */
	if (bes->mode & BES_VFILTER)
		intrep = (bes->srcheight == cropped_height) ? 0 : 1;
	else
		intrep = (bes->srcheight >= cropped_height) ? 0 : 1;
	ifactor = ((cropped_height - intrep) << 14) / (bes->dstheight - intrep);
	ifactorbetter = ((cropped_height - intrep) << 20) / (bes->dstheight - intrep) + intrep;
	condition = ( ((ifactorbetter * (bes->dstheight-1)) >> 20) > ((ifactor * (bes->dstheight-1)) >> 14) );
	if (condition)
		roundoff = !(bes->mode & BES_VFILTER) ? 1 : 0;
	else
		roundoff = 0;
	regs->besviscal = ifactor + roundoff;
	if (regs->besviscal >= (32 << 14))
		regs->besviscal = (32 << 14) - 1;

	/* Vertical subpixel compensation */
	if (regs->besviscal >= (1 << 14)) /* Downscaling */
		subpixelcomp = regs->besviscal - (1 << 13);
	else                        /* Upscaling */
		subpixelcomp = ((bes->dstheight - intrep) << 14)
		                / ((cropped_height - intrep) * 2);

	if (bes->frame == BES_FRAME_AUTO) {
		compute_vertical_source_positionning(bes, BES_FIELD_A1, 0);
		compute_vertical_source_positionning(bes, BES_FIELD_A2, subpixelcomp);
		compute_vertical_source_positionning(bes, BES_FIELD_B1, 0);
		compute_vertical_source_positionning(bes, BES_FIELD_B2, subpixelcomp);
	} else {
		compute_vertical_source_positionning(bes, bes->frame, 0);
	}
	
	regs->besviscal <<= 2;
}

static void mgabes_write_regs(struct mga_bes_s *bes)
{
	struct bes_registers_s *regs = &bes->regs;
	struct mga_dev_s *mga = bes->mga;

	/* Make sure internal registers don't get updated until we're done */
	mga_write(mga, BESGLOBCTL, (mga_read(mga, VCOUNT)-1)<<16);

	mga_write(mga, BESCTL, regs->besctl);
	mga_write(mga, BESPITCH, regs->bespitch); 

	mga_write(mga, BESA1ORG, regs->besa1org);
	mga_write(mga, BESA1CORG, regs->besa1corg);
	mga_write(mga, BESA2ORG, regs->besa2org);
	mga_write(mga, BESA2CORG, regs->besa2corg);
	mga_write(mga, BESB1ORG, regs->besb1org);
	mga_write(mga, BESB1CORG, regs->besb1corg);
	mga_write(mga, BESB2ORG, regs->besb2org);
	mga_write(mga, BESB2CORG, regs->besb2corg);

	if(bes->mode & BES_IS_G400) {
		mga_write(mga, BESLUMACTL, regs->beslumactl); 
		mga_write(mga, BESA1C3ORG, regs->besa1c3org);
		mga_write(mga, BESA2C3ORG, regs->besa2c3org);
		mga_write(mga, BESB1C3ORG, regs->besb1c3org);
		mga_write(mga, BESB2C3ORG, regs->besb2c3org);
	}
	mga_write(mga, BESHCOORD, regs->beshcoord);
	mga_write(mga, BESHISCAL, regs->beshiscal);
	mga_write(mga, BESHSRCST,  regs->beshsrcst);
	mga_write(mga, BESHSRCEND, regs->beshsrcend);
	mga_write(mga, BESHSRCLST, regs->beshsrclst);
	
	mga_write(mga, BESVCOORD, regs->besvcoord);
	mga_write(mga, BESVISCAL, regs->besviscal);
	mga_write(mga, BESV1SRCLST, regs->besv1srclst);
	mga_write(mga, BESV1WGHT, regs->besv1wght);
	
	/* update the registers somewhere between 1 and 2 frames from now. */
	mga_write(mga, BESGLOBCTL, regs->besglobctl + ((mga_read(mga, VCOUNT)+2)<<16));
}

/* see p. 4-96 in the G200 spec */
static void mgabes_program(struct mga_bes_s *bes)
{
	struct bes_registers_s *regs;
	u32 bes_was_enabled;

	regs = &bes->regs;

	/* window coordinates */
	regs->beshcoord = (bes->x + bes->dstwidth)  + (bes->x << 16);
	regs->besvcoord = (bes->y + bes->dstheight) + (bes->y << 16);

	regs->bespitch = mgabes_pitch_rounding(bes, bes->srcwidth);

	/* save this value, as we will overwrite it in the next subroutine call */
	bes_was_enabled = (bes->regs.besctl & 1);

	compute_horizontal_scaling(bes);
	compute_vertical_scaling(bes);

	/* 
	 * see p. 3-52 G200 specs
	 */
	regs->besctl |= bes_was_enabled
		+ ((bes->mode & BES_HFILTER) ? (1<<10) : 0)
		+ ((bes->mode & BES_VFILTER) ? (1<<11) : 0)
		+ ((bes->mode & BES_HFILTER) ? (1<<16) : 0)
		+ ((bes->mode & BES_DITHER) ? (1<<18) : 0)
		+ ((bes->mode & BES_BW) ? (1<<20) : 0)
		+ ((bes->frame & BES_FRAME_AUTO) ? (1<<24) : 0);

	if (!(bes->frame & BES_FRAME_AUTO))
		regs->besctl |= (bes->frame << 25);

	regs->besglobctl = ((bes->mode & BES_ACC2XZOOM) ? 1 : 0)
		| ((bes->mode & BES_HFILTER) ? 2 : 0);

	mgabes_write_regs(bes);
}

/**
 *	mgabes_activate -	activate the BES
 *	@bes:	bes instance
 *
 *	This function actives the BackEnd Scaler, so that the video
 *	overlay is displayed on screen.
 */
void mgabes_activate(struct mga_bes_s* bes)
{
	if (!bes) return;

        /* Activate scaler */
	bes->regs.besctl |= 1;
	mgabes_write_regs(bes);
}


/**
 *	mgabes_disable -	disables the BES
 *	@bes:	bes instance
 *
 *	This function stops the BackEnd Scaler, disabling the
 *	display of overlaid video on screen.
 */
void mgabes_desactivate(struct mga_bes_s* bes)
{
	if (!bes) return;

	bes->regs.besctl &= ~1;
	/*
	 * do not reprogram the whole register set,
	 * just stop the BES, because we're not sure
	 * that the parameters are still valid and won't
	 * crash the computation routines.
	 */
	mga_write(bes->mga, BESCTL, bes->regs.besctl);
}

/**
 *	mgabes_set_window -	change displayed video window
 *	@bes:	bes instance
 *	@x:	x coordinate of the overlay window
 *	@y:	y coordinate of the overlay window
 *	@width:	width of the overlay window
 *	@height: height of the overlay window
 *
 *	This function moves and changes the size of the overlay window.
 *
 *	Notes: The BES must be activated for this function to have a
 *	*visible* effect... However, the BES parameters are immediatly
 *	re-programmed.
 */
void mgabes_set_window(struct mga_bes_s *bes, int x, int y, int width, int height)
{
	if (!bes) return;
	
	bes->x = x;
	bes->y = y;
	bes->dstwidth = width;
	bes->dstheight = height;
	mgabes_program(bes);
}

/**
 *	mgabes_set_source -	set the source video buffer
 *	@bes:		bes instance
 *	@width:		width of a source video frame
 *	@height:	height of a source video frame
 *	@base_a:	base address of video buffer A
 *	@base_b:	base address of video buffer B
 *	@dataformat:	format of the video data
 *
 *	This function attaches the BES to one or two source video
 *	buffers, with the specified dimensions and data format.
 *	
 *	The 2nd buffer can be NULL. In this case, the Automatic Field
 *	Select mode shouldn't be used.
 *
 *	Formats:	%VIDEO_PALETTE_YUV420,
 *	%VIDEO_PALETTE_YUV420P, and %VIDEO_PALETTE_YUV422
 *
 *	Note:	This function will only have an effect after
 *	mgabes_enable() is called.
 *
 *	See also:	%VIDEO_PALETTE_ and struct &video_buffer in
 *	<videodev.h> 
 */
void mgabes_set_source(struct mga_bes_s *bes, int width, int height,
		       u32 base_a, u32 base_b, int dataformat)
{

	bes->srcwidth = width;
	bes->srcheight = height;
	bes->buf_a = base_a;
	bes->buf_b = base_b;
	/* FIXME: coherence palette <-> dataformat ! */
	bes->palette = dataformat;
}

/**
 *	mgabes_set_colorkey -	change overlay colorkey
 *	@bes:		bes instance
 *	@chromakey:	new chromakey
 *	
 *	This function immediatly updates the chromakey used for
 *	the overlay window.
 */
void mgabes_set_colorkey(struct mga_bes_s *bes, u32 chromakey)
{
	struct mga_dev_s *mga;
	int r, g, b;

	bes->chromakey = chromakey;
	r = (bes->chromakey >> 24) & 0xFF;
	g = (bes->chromakey >> 16) & 0xFF;
	b = (bes->chromakey >> 8) & 0xFF;

	mga = bes->mga;

	/* set colorkey */
	switch (mga_indirect_readb(mga, XMULCTRL)) {
	case BPP_8:
		/* Need to look up the color index, just using
		   color 0 for now. */
		break;
	case BPP_15:
		r >>= 3;
		g >>= 3;
		b >>= 3;
		break;
	case BPP_16:
		r >>= 3;
		g >>= 2;
		b >>= 3;
		break;
	case BPP_24:
	case BPP_32_DIR:
	case BPP_32_PAL:
	default:
		break;
	}

	mga_indirect_writeb(mga, XKEYOPMODE, 1);
	mga_indirect_writeb(mga, XCOLMSK0RED, 0xff);
	mga_indirect_writeb(mga, XCOLMSK0GREEN, 0xff);
	mga_indirect_writeb(mga, XCOLMSK0BLUE, 0xff);
	mga_indirect_writeb(mga, XCOLKEY0RED, r);
	mga_indirect_writeb(mga, XCOLKEY0GREEN, g);
	mga_indirect_writeb(mga, XCOLKEY0BLUE, b);
}

/**
 *	mgabes_set_frame -	change the displayed frame
 *	@bes:	bes instance
 *	@frame:	frame to display
 *
 *	mga_set_frame() changes the frame displayed in the overlay
 *	window. The 'frame' parameter can be %BES_FRAME_A or
 *	%BES_FRAME_B to toggle between one of the 2 buffers. The
 *	%BES_FRAME_AUTO parameter should be used in conjunction with
 *	the video grabber to automagically preview the captured video
 *	stream.
 *
 *	Notes:	This function reprograms the BES immediatly. You
 *	should make sure that valid parameters have been entered
 *	previously (like video source buffers or video window).
 *
 *	See also:	4.10.5.2 Automatic Field Select p. 4-111 of
 *	the G200 spec.
 */
void mgabes_set_frame(struct mga_bes_s *bes, int frame)
{
	bes->frame = frame;
	mgabes_program(bes);
}

/**
 *	mgabes_set_mode -	change BES mode/options
 *	@bes:	bes instance
 *	@mode:	new mode
 *
 *	mga_bes_set_mode() changes the operating mode and or options
 *	of the BES.
 *
 *	Modes:
 *	BES_HFILTER	horizontal filtering
 *
 *	BES_VFILTER	vertical filterting
 *
 *	BES_ACC2XZOOM	2X Zoom (for high MCLKs)
 *
 *	BES_DITHER	dither enabled
 *
 *	BES_BW		black & white mode
 *
 *	BES_IS_G400	watch out for G400 specific parameters (needs
 *	more support)
 *
 *	BES_INTERLACED	source video data is interlaced
 *
 *	Note:	This function will only have an effect after
 *	mgabes_enable() is called.
 *
 */
void mgabes_set_mode(struct mga_bes_s *bes, int mode)
{
	bes->mode = mode;
}
