/*
 * This module is an interface to the ZORAN ZR367x0 DVD decoder chip,
 * also known as I34.
 *
 * (c) 1999-2000 David Barth <dbarth@besancon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <linux/videodev.h>
#include "mgacore.h"
#include "mgacodec.h"
#include "mgagrab.h"
#include "mgabes.h"
#include "i34.h"

int debug = 1; 	 	/* insmod parameter */

MODULE_AUTHOR("David Barth <dbarth@besancon.net>");
MODULE_DESCRIPTION("Driver for the Matrox DVD module (ZR36700)");
MODULE_PARM(debug,"i");

static struct mga_dev_s *mga;
static int in_use;
static u32 codecbuffer;
static u32 videobuffer;

/****************************************************************************
 * V4L1 INTERFACE                                                           *
 ****************************************************************************/

static long mgadvd_write(struct video_device *dev, const char *buf, unsigned long count, int noblock)
{
	unsigned long todo = count;
	unsigned int  chunk;
	int ret;

	while (todo > 0) {
		chunk = (todo > CODEC_BUFFER_SIZE/4) ? CODEC_BUFFER_SIZE/4 : todo;
		if ((ret = mgacodec_send_bitstream(mga->codec, buf, chunk)) < 0)
			return (long)ret;
		buf  += chunk;
		todo -= chunk;
	}
	return count;
}

static long mgadvd_read(struct video_device *dev, char *buf, unsigned long count, int noblock)
{
	return -EINVAL;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
unsigned int mgadvd_poll(struct video_device *dev, struct file *file, struct poll_table_struct *wait)
#else
unsigned int mgadvd_poll(struct file *file, struct poll_table_struct *wait)
#endif
{
	unsigned int mask = 0;

	/* if (!&host_mga.transfer) return POLLOUT | POLLWRNORM; */
	/* if (mgacodec_buffer_free(&host_mga) > 0)
	   return mask = POLLOUT | POLLWRNORM;
	*/
	/* poll_wait(file, &host_mga.codec_inq, wait); */
	/* if (mgacodec_buffer_free(&host_mga) > 0)
		mask = POLLOUT | POLLWRNORM;
	*/
	return mask;
}


static int mgadvd_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	int ret;

	switch(cmd) {

		/* first the official IOCTLs */
	case VIDIOCGCAP:
	{	
		struct video_capability cap;
		cap.type = VID_TYPE_MPEG_DECODER | VID_TYPE_FRAMERAM | VID_TYPE_SCALES;
		cap.channels = 0;
		cap.audios = 1;
		cap.maxwidth = 0;
		cap.maxheight = 0;
		cap.minwidth = 0;
		cap.minheight = 0;
		strcpy(cap.name, MGADVD_ID_STRING);
		if (copy_to_user(arg, &cap, sizeof(cap)))
			return -EFAULT;
		return 0;
	}
	case VIDIOCGFBUF:
		return -ENOIOCTLCMD;
		break;

	case VIDIOCSFBUF:
		return -ENOIOCTLCMD;
		break;

	case VIDIOCGAUDIO:
		return -ENOIOCTLCMD;
		break;

	case VIDIOCSAUDIO:
		return -ENOIOCTLCMD;
		break;

	case VIDIOCGUNIT:
	{
		struct video_unit vu;
		vu.video = dev->minor;
		vu.vbi = VIDEO_NO_UNIT;
		vu.radio = VIDEO_NO_UNIT;
		vu.audio = VIDEO_NO_UNIT;
		vu.teletext = VIDEO_NO_UNIT;
		if (copy_to_user(arg, &vu, sizeof(vu)))
			return -EFAULT;

		return 0;
	}
	case VIDIOCSWIN:
	{
		struct video_window vw;

		if (copy_from_user(&vw, arg, sizeof(vw)))
			return -EFAULT;
		if (vw.x <0 || vw.y <0 || vw.width < 32 
		   || vw.height < 16)
			return -EINVAL;
		if (vw.flags == VIDEO_WINDOW_CHROMAKEY)
			mgabes_set_colorkey(mga->bes, vw.chromakey);
		mgabes_set_window(mga->bes, vw.x, vw.y, vw.width, vw.height);

		return 0;
	}
	case VIDIOCGPICT:
	{
		/*
		 *	Default values.. if we can change this we
		 *	can add the feature later
		 */
		struct video_picture vp;

		vp.brightness = 0x8000;
		vp.hue = 0x8000;
		vp.colour = 0x8000;
		vp.whiteness = 0x8000;
		vp.depth = 8;
		vp.palette = VIDEO_PALETTE_YUV420;
		if(copy_to_user(arg, &vp, sizeof(vp)))
			return -EFAULT;
		return 0;
	}
	case VIDIOCGWIN:
	{
		return -ENOIOCTLCMD;
	}
	case VIDIOCSPICT:
	{
		return -ENOIOCTLCMD;
	}
	case VIDIOCCAPTURE:
	{
		if (arg)
			mgabes_activate(mga->bes);
		else
			mgabes_desactivate(mga->bes);
		return 0;
	}
	
	case VIDIOCSPLAYMODE:
	{
		struct video_play_mode pmode;
		if (copy_from_user((void *)&pmode, arg,
				   sizeof(struct video_play_mode)))
			return -EFAULT;
		switch (pmode.mode)
		{
		case VID_PLAY_VID_OUT_MODE:
			if (pmode.p1 != VIDEO_MODE_NTSC &&
			    pmode.p1 != VIDEO_MODE_PAL)
				return -EINVAL;
			if (!i34_video_standard(pmode.p1))
				return -EBUSY;
			break;

		case VID_PLAY_NORMAL:
			if (!i34_start())
				return -EBUSY;
			break;

		case VID_PLAY_PAUSE:
			if (!i34_pause())
				return -EBUSY;
			break;

		case VID_PLAY_RESET:
			if (!i34_init()) return -EBUSY;
			break;

		case VID_PLAY_IMMEDIATE_NORMAL:
			if (!i34_continue()) return -EBUSY;
			break;

		case VID_PLAY_MASTER_MODE:
			
			break;

		case VID_PLAY_GENLOCK:
		case VID_PLAY_SINGLE_FRAME:
		case VID_PLAY_FAST_FORWARD:
		case VID_PLAY_SLOW_MOTION:
		case VID_PLAY_SWITCH_CHANNELS:
		case VID_PLAY_FREEZE_FRAME:
		case VID_PLAY_STILL_MODE:
		case VID_PLAY_ACTIVE_SCANLINES:
		case VID_PLAY_END_MARK:
		default:
			return -EINVAL;
		}
		return 0;
	}

	case VIDIOCSMICROCODE:
	{
		struct video_code ucode;

		if (copy_from_user((void *) &ucode, arg, sizeof(ucode)))
			return -EFAULT;
		if (ucode.datasize > 16384) /* Zoran ucode is small... */
			return -EINVAL;
		if ((ucode.data = vmalloc(ucode.datasize)) == NULL)
			return -ENOMEM;
		if (copy_from_user((void *) ucode.data, ucode.data, ucode.datasize)) {
			vfree(ucode.data);
			return -EFAULT;
		}

		/* return i34_install_microcode(ucode); */
		return -EINVAL;
	}

	/* then the private ones */

	case I34_IOCGSTATUS:
	{
		struct mgadvd_status_s status;

		if ((ret = i34_status(&status)))
			return ret;
		if (copy_to_user((struct mgadvd_status_s *)arg, &status,
		     sizeof(struct mgadvd_status_s)))
			return -EFAULT;
	}

	case I34_IOCSVIDEOSID:
		if (!i34_change_video_sid((int)arg))
			return -EINVAL;
		break;

	case I34_IOCSAUDIOSID:
		if (!i34_change_audio_sid((int)arg))
			return -EINVAL;
		break;

	case I34_IOCSSUBPICSID:
		if (!i34_change_subpic_sid((int)arg))
               		return -EINVAL;
		break;

	case I34_IOCSADPMICROCODE:
		if (!i34_adp_boot((int)arg))
			return -EINVAL;
		break;

	case I34_IOCSDVPMICROCODE:
		if (!i34_dvp_boot((int)arg))
			return -EINVAL;
		break;

	case I34_IOCSSTREAMTYPE:
		if (!i34_select_bitstream((int)arg))
			return -EINVAL;
		break;

	case I34_IOCSVIDEOSYNC:
		if (!i34_video_sync((int)arg))
			return -EINVAL;
		break;

	case I34_IOCSAUDIOSYNC:
		if (!i34_audio_sync((int)arg))
			return -EINVAL;
		break;

#ifdef CSS_SUPPORT
	case I34_IOCS_DISK_KEY:
		if (copy_from_user(&i34.cssauth, (dvd_authinfo *)arg, sizeof(dvd_authinfo)))
			return -EFAULT;
		if (!css_write_disk_key(i34.cssauth))
			return -EINVAL;
		break;

	case I34_IOCS_TITLE_KEY:
		if (copy_from_user(&i34.cssauth, (dvd_authinfo *)arg, sizeof(dvd_authinfo)))
			return -EFAULT;
		if (!css_write_title_key(i34.cssauth))
			return -EINVAL;
		break;

	case I34_IOCG_CHAL_KEY:
		if (!css_read_chal_key(i34.cssauth))
			return -EINVAL;
		if (copy_to_user((dvd_authinfo *)arg, &i34.cssauth, sizeof(dvd_authinfo)))
			return -EFAULT;
		break;

	case I34_IOCS_CHAL_KEY:
		if (copy_from_user(&i34.cssauth, (dvd_authinfo *)arg, sizeof(dvd_authinfo)))
			return -EFAULT;
		if (!css_write_chal_key(i34.cssauth))
			return -EINVAL;
		break;

	case I34_IOCG_RESP_KEY:
		if (!css_read_resp_key(i34.cssauth))
			return -EINVAL;
		if (copy_to_user((dvd_authinfo *)arg, &i34.cssauth, sizeof(dvd_authinfo)))
			return -EFAULT;
		break;

	case I34_IOCS_RESP_KEY:
		if (copy_from_user(&i34.cssauth, (dvd_authinfo *)arg, sizeof(dvd_authinfo)))
			return -EFAULT;
		if (!css_write_resp_key(i34.cssauth))
			return -EINVAL;
		break;
#endif

	default:    
		return -ENOIOCTLCMD;
	}

	return 0;
}

static inline void mga_draw_box(struct mga_dev_s *mga, int x, int color)
{
	int i,j;

	for(i=0; i<20; ++i)
		for(j=x; j<x+20; ++j)
			mga->fb[i*1280*2+j] = color;
}

static void vsync_debug(struct mga_dev_s *mga, struct pt_regs *regs)
{
	u16 s0, s1;

	writel(1, mga->ctrl + VICLEAR);

	/*
	  i34_write_parameter(0x48, 0);
	  i34_write_parameter(0x48, 3);
	  i34_read_sclk();
	*/

	s0 = i34_read(0x3);
	s1 = i34_read(0x4);

	if (s0 & 0x40) { /* Vidbuf empty, square left top */
		mga_draw_box(mga, 0, 0xff);
	} else {
		mga_draw_box(mga, 0, 0);
	}
	if (s0 & 0x20) { /* Vidbuf full, square in middle */
		mga_draw_box(mga, 30, 0xff);
	} else {
		mga_draw_box(mga, 30, 0);
	}

	if (s1 & 0x4000) { /* starving or SCLK behind DTS */
		mga_draw_box(mga, 60, 0xff);
	} else {
		mga_draw_box(mga, 60, 0);
	}
}

static int mgadvd_open(struct video_device *dev, int flags)
{
	int error;
	error = 0;

	if (in_use)
		return -EBUSY;

	mga_request_irq(mga);

	// FIXME: the interrupt handler is broken now
	// here's what needs to be done
	// mga_request_irq(mga)
	//    where it clears all MGA interrupt sources and
	//    installs a handler for the PCI interrupt
	//    (init w/ all MGA interrupts OFF)
	// mga_select_irq(MGA_IRQ_VSYNC, vsync_debug)
	//    where it selects which MGA interrupt and handler to use
	//    *** NOT VINVSYNC because that NEVER triggers with the 36710
	//    vsync_debug does NOT belong here. move it to core and
	//    fix it so it will use the screen size in mga structure
	//    instead of assuming 1280 width, which is a BAD assumption
	//
	// I think these HAVE to be broken apart this way.
	// I'll try to fix it, but it's a REAL BAD MESS
	// -inkling
	
	if (!i34_init())
		/* FIXME: EBUSY or ENODEV ? */
		return -EBUSY;

	mgabes_set_mode(mga->bes, BES_HFILTER | BES_VFILTER | BES_ACC2XZOOM | BES_DITHER);
	mgabes_set_source(mga->bes, 720, 576,
			  videobuffer, videobuffer, /* frame B = frame A */
			  VIDEO_PALETTE_YUV422);
	mgabes_set_frame(mga->bes, BES_FRAME_AUTO);

	mgagrab_activate(mga->grabber, GRAB_AUTO);

	in_use++;

	/* FIXME: MOD_INC_USE_COUNT; */
	return 0;
}

static void mgadvd_close(struct video_device *dev)
{
	/* clean up things before releasing the device */

	i34_stop();
	mga_release_irq(mga);
	mgagrab_desactivate(mga->grabber);
	mgabes_desactivate(mga->bes);

	in_use--;
	/* FIXME: MOD_DEC_USE_COUNT; */
}


static struct video_device mgadvd_video_device =
{
	MGADVD_ID_STRING,
	VID_TYPE_MPEG_DECODER | VID_TYPE_FRAMERAM | VID_TYPE_SCALES,
	VID_HARDWARE_ZR36700,
	mgadvd_open,
	mgadvd_close,
	mgadvd_read,
	mgadvd_write,
	mgadvd_poll,
	mgadvd_ioctl,
	NULL,
	NULL,
	NULL
};

#ifdef MODULE
/* hide all global symbols */
EXPORT_NO_SYMBOLS;
int init_module(void)
#else
int mgadvd_init(void)
#endif
{
	int status;

	status = 0;
	codecbuffer = 0;
	videobuffer = 0;

	printk("Matrox DVD interface v%s (c) David Barth\n", MGADVD_VERSION);

	mga = mga_get_card();
	if (!mga)
	{
		printk("mgadvd: no supported host card found\n");
		goto FAIL;
	}

	/* allocate a buffer for the codec interface
	   and add a security margin (0x400 bytes)
	   for further alignment requirement.
	*/
	codecbuffer = (u32)mga_alloc_buffer(mga, CODEC_BUFFER_SIZE + 0x400);
	mgacodec_init(mga, codecbuffer, MGACODEC_MODE_VMI_A, ZR_BITSTREAM);

	if (i34_detect(mga->codec) < 0) {
		printk("mgadvd: no decoder card detected\n");
		status = -ENODEV;
		goto FAIL;
	}

	mgadvd_video_device.priv = mga;

	if (video_register_device(&mgadvd_video_device, VFL_TYPE_GRABBER) == -1) {
		printk("mgadvd: unable to register video device\n");
		status = -EINVAL;
		goto FAIL;
	}

	videobuffer = (u32)mga_alloc_buffer(mga, 720 * 576 * 2 + 0x100);
	mgagrab_init(mga, 720, 576, videobuffer, videobuffer);

	mgabes_init(mga);

	return 0;

 FAIL:
	if (codecbuffer && mga)
		mga_free_buffer(mga, CODEC_BUFFER_SIZE + 0x400);
	if (videobuffer && mga)
		mga_free_buffer(mga, 720 * 576 * 2 + 0x100);
	if (mga)
		mga_release_card(mga);

	return status;
}

#ifdef MODULE
void cleanup_module(void)
{
	printk("mgadvd: cleaning up module\n");
	video_unregister_device(&mgadvd_video_device);
	mga_free_buffer(mga, CODEC_BUFFER_SIZE + 0x400);
	mga_free_buffer(mga, 720 * 576 * 2 + 0x100);
	mga_release_card(mga);
}
#endif
