/*
 * The current driver is modeled after the G200 specs
 * There are many more features available in G400 mode
 *
 * see p. 4.61 (459) in the G200 specs
 */

#define __NO_VERSION__
#include <linux/module.h>
#include "mgacore.h"
#include "mgagrab.h"

#define VINADDR0        0x3e10
#define VINADDR1        0x3e14
#define VINCTL          0x3e1c
#define VINCTL0         0x3e00
#define VINCTL1         0x3e04
#define VINNEXTWIN      0x3e18
#define VBIADDR0        0x3e08
#define VBIADDR1        0x3e0c

struct vin_regs_s
{
	u32 vbiaddr0;
	u32 vbiaddr1;
	u32 vinaddr0;
	u32 vinaddr1;
	u32 vinctl;
	u32 vinctl0;
	u32 vinctl1;
	u32 vinnextwin;
};

struct mga_grab_s
{
	int srcwidth;
	int srcheight;	/* unused with G200 */
	int mode;
	u32 window0;
	u32 window1;
	struct vin_regs_s regs;
	struct mga_dev_s *mga;
};

static struct mga_grab_s unique_grabber;

static int mgagrab_write_regs(struct mga_grab_s *grabber) {
	struct mga_dev_s *mga;
	struct vin_regs_s *regs;

	mga = grabber->mga;
	regs = &grabber->regs;

	/* to update the registers we must make sure that we're not accessing
	   a window that is used (see 4-82 in the G200 spec)
	*/

	mga_write(mga, VINCTL, regs->vinctl);
	mga_write(mga, VBIADDR0, regs->vbiaddr0);
	mga_write(mga, VINADDR0, regs->vinaddr0);
	mga_write(mga, VINCTL0, regs->vinctl0);
	mga_write(mga, VBIADDR1, regs->vbiaddr1);
	mga_write(mga, VINADDR1, regs->vinaddr1);
	mga_write(mga, VINCTL1, regs->vinctl1);
	mga_write(mga, VINNEXTWIN, regs->vinnextwin);
	/* FIXME: the G400 works differently */
	/* if (mga->is_g400) { */
		/* vinheight */
		/* vinaddr2-3, vinctl2-3 */
	/* } */

	return 0;
}


static void mgagrab_program(struct mga_grab_s *grabber)
{
	struct mga_dev_s *mga;
	struct vin_regs_s *regs;
	u32 vinpitch;

	mga = grabber->mga;
	regs = &grabber->regs;

	/* see p. 4-111 in the G200 spec */
	/* Step 1: Set the starting field for the window
	   (besfselm = 0 at this time)
	   besfselm = 0;
	   regs.besctl &= ~(besfselm<<24);
	   mga_vid_write_regs();
	   Step 2: switch to hardware field select
	   besfselm = 1;
	   regs.besctl |= (besfselm<<24);
	   mga_vid_write_regs();
	*/
  
	/* see p. 4-81 of the G200 spec */
	/* Now, program the video grabber so that it will fill
	   the BES buffers with video input */
  
	vinpitch = (grabber->srcwidth * 2) & ~3;
	regs->vinaddr0 = grabber->window0 & ~7;
	regs->vinaddr1 = (grabber->window1 + vinpitch) & ~7;
	regs->vinctl0 = (1 | 00 | ((vinpitch>>2)<<3)) & 0x0FFFF; /* no VBI capture */	
	regs->vinctl1 = (1 | 00 | ((vinpitch>>2)<<3)) & 0x0FFFF; /* no VBI capture */	
	regs->vinnextwin = 3; 
	/* FIXME: the G400 works differently */

	regs->vinctl = (grabber->mode) ? 1 : 0;

	/* FIXME remove */
	/* printk("vinaddr0 = %#08lx\n", regs->vinaddr0);
	   printk("vinaddr1 = %#08lx\n", regs->vinaddr1);
	   printk("vinpitch = %#04x\n", vinpitch);
	*/

	mgagrab_write_regs(grabber);
}


/****************************************************************************
 * PUBLIC INTERFACE                                                         *
 *                                                                          *
 ****************************************************************************/

EXPORT_SYMBOL(mgagrab_init);
EXPORT_SYMBOL(mgagrab_activate);
EXPORT_SYMBOL(mgagrab_desactivate);

/**
 *	mgagrab_init - 
 **/
int mgagrab_init(struct mga_dev_s *mga, int width, int height, u32 win0, u32 win1)
{
	struct mga_grab_s *grabber;

	grabber = &unique_grabber;
	mga->grabber = grabber;
	grabber->mga = mga;
	grabber->srcwidth = width;
	grabber->srcheight = height;
	grabber->window0 = win0;
	grabber->window1 = win1;

	return 0;
}

/**
 *	mgagrab_activate - 
 **/
int mgagrab_activate(struct mga_grab_s *grabber, int mode)
{
	grabber->mode = mode;
	mgagrab_program(grabber);

	return 0;
}

/**
 *	mgagrab_desactivate -
 **/
int mgagrab_desactivate(struct mga_grab_s *grabber)
{
	grabber->mode = 0;
	mgagrab_program(grabber);

	return 0;
}
