
/*
 *
 * Copyright (C) 1999  Thomas Mirlacher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author may be reached as dent@cosy.sbg.ac.at, or
 * Thomas Mirlacher, Jakob-Haringerstr. 2, A-5020 Salzburg,
 * Austria
 *
 *------------------------------------------------------------
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <dvb/ts.h>
#include <dvb/pes.h>
#include <oms/oms.h>
#include <oms/buf.h>
#include <oms/plugin/codec.h>

static int _ts_init (plugin_codec_t *plugin, buf_t *buf, char *output_path);
static int _ts_open (plugin_codec_t *plugin);
static int _ts_read (plugin_codec_t *plugin, buf_t *buf, buf_entry_t *buf_entry);
static int _ts_set_stat (plugin_codec_t *plugin, u_int status);
static int _ts_set_attr (plugin_codec_t *plugin, u_int id, void *attr);


static struct {
	buf_t *buf;

	int fd;
	int pid;
} _ts_priv;

static plugin_codec_t codec_sw = {
	_ts_init,
	NULL,
	_ts_open,
	NULL,
	_ts_read,	// read
	NULL,		// write
	_ts_set_stat,
	_ts_set_attr,
}; 

static int ts_encaps (buf_entry_t *buf_entry, const int pid);


/**
 *
 **/

static int _ts_init (plugin_codec_t *plugin, buf_t *buf, char *output_path)
{
	_ts_priv.buf = buf;
	
	_ts_priv.fd = open (output_path, O_WRONLY | O_CREAT);
	_ts_priv.pid = 0xa0;

	return 0;
}


/**
 *
 **/

static int _ts_set_stat (plugin_codec_t *plugin, u_int status)
{
	return 0;
}


/**
 * Set attributes
 **/

static int _ts_set_attr (plugin_codec_t *plugin, u_int id, void *attr)
{
	return 0;
}


// DENT: eliminate this ...
static char cont_count[0x1FFF]={0,};
static char ts_buf[TS_LEN*50];	// max 20 packets

// DENT: the implementation has been simple ported from my stand-alone
//	ps2ts program, so it's SOMEWHAT sub optimal here.
//	since we' like to output the data to a device or file in any case,
//	there's no need to put the data into a ts_buf first -> change this
//	favour of iovs

 
/**
 *
 **/

static int _ts_read (plugin_codec_t *plugin, buf_t *buf, buf_entry_t *buf_entry)
{
	int pusi = 0x01;
	int data_txed = 0;
	int packets = 0;
	int pid = _ts_priv.pid;

	char *inp = buf_entry->data;
	int len = buf_entry->data_len;
	char *outp = &ts_buf;
	long long clock = buf_entry->clock;
	long clock_ext = buf_entry->clock_ext;
	int set_clock = buf_entry->flags & BUF_FLAG_CLOCK_VALID;

// DENT: the buffer contains the mem_start of something (PES, PS, ..),
//	but we can be sure it contains data_start
//	so it would be an overhead to add an extra pointer, pointing
//	to the start of the PES hdr.
//	since we can be sure, there's enough room before the data for
//	a minimal PES hdr - just add one ;)

// add PES hdr
	pes_hdr_t *pes;
	inp -= PES_HDR_LEN + PES_FLAG_LEN;
	pes = (pes_hdr_t *) inp;

	pes->start_code_prefix = PES_START_CODE_PREFIX_RAW;
	pes->stream_id = 0xE0;
	pes->pes_packet_length_hi = (len>>8) & 0xFF;
	pes->pes_packet_length_lo = len & 0xFF;

	*(inp+PES_HDR_LEN) = 0;	// set flags, and pes_header_data_length
	*(inp+PES_HDR_LEN+1) = 0;	// set flags, and pes_header_data_length
	*(inp+PES_HDR_LEN+2) = 0;	// 
	len += PES_HDR_LEN + PES_FLAG_LEN;
// done PES hdr

	while (data_txed < len) {
		int data_len = (len-data_txed-TS_DATA_LEN) >= 0 ? TS_DATA_LEN : len-data_txed;

		ts_hdr_t *ts_head = (ts_hdr_t *) outp;
		outp += TS_HDR_LEN;
		
		if (set_clock && (data_len > TS_DATA_LEN-TS_ADAPTATION_FIELD_LEN-TS_PCR_LEN-1))
			data_len -= TS_ADAPTATION_FIELD_LEN + TS_PCR_LEN + 1;// -1 for sizeof (adaptation_field_length) byte
	
		ts_head->sync_byte = TS_SYNC_BYTE;
		ts_head->transport_error_indicator = 0;
		ts_head->payload_unit_start_indicator = pusi;
		ts_head->transport_priority = 0;
		ts_head->PID_hi = pid >> 8;
		ts_head->PID_lo = pid & 0xFF;
		
		ts_head->continuity_counter = cont_count[pid];
		ts_head->transport_scrambling_control = 0;
		
		
		if (data_len != TS_DATA_LEN) {				// we've to add an adaptation field here
			int adapt_len = TS_DATA_LEN - data_len - 1;

			ts_head->adaptation_field_control = 3;		// adaptation field followed by payload
			
			*outp++ = adapt_len;				// adaptation field length

			if (adapt_len) {				// we've at least flags to put in
				ts_adapt_fld_t *ts_adapt = (ts_adapt_fld_t *) outp;
				outp += TS_ADAPTATION_FIELD_LEN;
				adapt_len -= TS_ADAPTATION_FIELD_LEN;
				
				ts_adapt->adaptation_field_extension_flag = 0;
				ts_adapt->transport_private_data_flag = 0;
				ts_adapt->splicing_point_flag = 0;
				ts_adapt->opcr_flag = 0;
				ts_adapt->pcr_flag = set_clock ? 1 : 0;
				ts_adapt->elementary_stream_priority_indicator = 0;
				ts_adapt->random_access_indicator = 0;
				ts_adapt->discontinuity_indicator = 0;	

				if (ts_adapt->pcr_flag) {		// fill in pcr
					ts_pcr_t *ts_pcr = (ts_pcr_t *) outp;
					outp += TS_PCR_LEN;
					adapt_len -= TS_PCR_LEN;

					ts_pcr->program_clock_reference_base_5 = clock;
					ts_pcr->program_clock_reference_base_4 = clock >>= 1;
					ts_pcr->program_clock_reference_base_3 = clock >>= 8;
					ts_pcr->program_clock_reference_base_2 = clock >>= 8;
					ts_pcr->program_clock_reference_base_1 = clock >>= 8;

					ts_pcr->program_clock_reference_extension_hi = clock_ext >> 8;
					ts_pcr->program_clock_reference_extension_lo = clock_ext; 

					set_clock = 0;
				}

				memset (outp, 0xFF, adapt_len);
				outp += adapt_len;
			}
		} else 
			ts_head->adaptation_field_control = 1;
		
		// ok, we're done with the header, so just fill in some data bytes
		memcpy (outp, inp+data_txed, data_len);
					
		outp += data_len;
		data_txed += data_len;
		pusi = 0;	
		
		packets++;
		
		cont_count[pid]++;
		cont_count[pid] %= 16;
	}

	write (_ts_priv.fd, ts_buf, packets*TS_LEN);
	return 0;
}


/**
 * Initialize Plugin.
 **/

void *plugin_init (char *whoami)
{
	pluginRegister (whoami,
		PLUGIN_ID_CODEC_VIDEO,
		STREAM_ID_MPEG2,
		&codec_sw);

	return &codec_sw;
}


/**
 * Cleanup Plugin.
 **/

void plugin_exit (void)
{
}
